from rasterio import features
import rasterio
from shapely.geometry import shape, mapping, Polygon, MultiPolygon, GeometryCollection
from skimage import measure
from skimage.segmentation import watershed
from skimage.feature import peak_local_max
from scipy import ndimage as ndi
import numpy as np
from typing import List, Optional, Union
from ml4floods.data.create_gt import get_brightness, BRIGHTNESS_THRESHOLD
from ml4floods.data import utils
import torch
import geopandas as gpd
from shapely.ops import unary_union
from shapely import validation
import pandas as pd
import warnings
from tqdm import tqdm
from datetime import datetime
import os


def _watershed_processing(prob_water_mask, thres_h=0.6, thres_l=0.4, distance= 5, conn=2, watershed_line=True):
    """
    Ancilliary function used to apply the watershed algorithm (to go from river lines river polygons). It doesn't
    work very well but maybe worth keeping in case it's useful in the future.

    Args:
        prob_water_mask:
        thres_h:
        thres_l:
        distance:
        conn:
        watershed_line:

    Returns:

    """
    mask0 = prob_water_mask > thres_h
    local_maxi = peak_local_max(prob_water_mask, indices=False, footprint=np.ones((distance * 2 + 1, distance * 2 + 1)),
                                labels=(prob_water_mask > thres_l))
    local_maxi[mask0] = True
    seed_msk = ndi.label(local_maxi)[0]

    prob_water_mask = watershed(-prob_water_mask, seed_msk, mask=(prob_water_mask > thres_l), watershed_line=watershed_line)
    return measure.label(prob_water_mask, connectivity=conn, background=0).astype('uint8')


def get_water_polygons(binary_water_mask: np.ndarray, min_area:float=25.5,
                       polygon_buffer:int=0, tolerance:float=1., transform: Optional[rasterio.Affine]=None) -> List[Polygon]:
    """
    Vectorize a binary mask excluding polygons smaller than `min_area` pixels.

    Args:
        binary_water_mask: (H, W) binary mask to rasterise
        min_area: polygons with pixel area lower than this will be filtered
        polygon_buffer: buffering of the polygons
        tolerance: to simplify the polygons, in pixel units
        transform: affine transformation of the binary_water_mask raster

    Returns:
        list of rasterised polygons

    """
    assert binary_water_mask.ndim == 2, f"Expected mask with 2 dim found {binary_water_mask.shape}"

    geoms_polygons = []
    polygon_generator = features.shapes(binary_water_mask.astype(np.int16),
                                        binary_water_mask)

    for polygon, value in polygon_generator:
        p = shape(polygon)
        if polygon_buffer > 0:
            p = p.buffer(polygon_buffer)
        if p.area >= min_area:
            p = p.simplify(tolerance=tolerance)
            if transform is not None:
                p = transform_polygon(p, transform) # Convert polygon to raster coordinates
            geoms_polygons.append(p)

    return geoms_polygons

def transform_polygon(polygon:Polygon, transform: rasterio.Affine) -> Polygon:
    """
    Transforms a polygon from pixel coordinates to the coordinates specified by the affine transform

    Args:
        polygon: polygon to transform
        transform: Affine transformation

    Returns:
        polygon with coordinates transformed by the affine transformation

    """
    geojson_dict = mapping(polygon)
    out_coords = []
    for pol in geojson_dict["coordinates"]:
        pol_out = []
        for coords in pol:
            pol_out.append(transform * coords)

        out_coords.append(pol_out)

    geojson_dict["coordinates"] = out_coords

    return shape(geojson_dict)


def get_mask_watertypes(mndwi: Union[np.ndarray, torch.Tensor],
                        water_mask:Union[np.ndarray, torch.Tensor],
                        permanent_water:Optional[Union[np.ndarray, torch.Tensor]]=None) -> Union[np.ndarray, torch.Tensor]:
    """
    This function returns a single mask that differenciates flood_water from permanent_water and flood traces.
    We distinguish flood traces using the mndwi index.

    Args:
        mndwi: (H, W) tensor MNDWI index
        water_mask:  (H, W) tensor with output classes provided by the model {0:invalid, 1: land, 2:water, 3:thick cloud}
        permanent_water: (H, W) tensor of the JRC permanent water layer coded as:
         https://developers.google.com/earth-engine/datasets/catalog/JRC_GSW1_3_YearlyHistory

    Returns:
        Water mask (H, W) with interpretation {0: invalids, 1: land, 2: flood water, 3: thick cloud, 4:permanent water, 5: flood trace}
    """
    # 2: flood water
    # 4: permanent_water
    # 5: flood_trace
    water_mask_types = water_mask.copy()
    water_mask_types[(water_mask == 2) & (mndwi < 0)] = 5
    if permanent_water is not None:
        water_mask_types[(water_mask != 3) & (permanent_water == 3)] = 4

    return water_mask_types


def get_pred_mask_v2(inputs: Union[np.ndarray, torch.Tensor], prediction: Union[np.ndarray, torch.Tensor],
                     channels_input:Optional[List[int]]=None,
                     th_water:float = 0.5, th_cloud:float = 0.5, mask_clouds:bool = True,
                     th_brightness:float=BRIGHTNESS_THRESHOLD,
                     collection_name:str="S2") -> Union[np.ndarray, torch.Tensor]:
    """
    Receives an output of a WFV2 model (multioutput binary) and returns the corresponding 3-class segmentation mask

    Args:
        inputs: S2 image (C, H, W)
        prediction: corresponding model output (2, H, W)
        channels_input: 0-based list of indexes of s2_image channels (expected that len(channels_input) == s2_image.shape[0])
        th_water: threshold for the class water
        th_cloud: threshold for the class cloud
        th_brightness: threshold for the brightness to differenciate thick from thin clouds
        mask_clouds: If False ignores brightness and outputs the prediction mask with thin clouds
        collection_name:
        
    Returns:
        Water mask (H, W) with interpretation {0: invalids, 1: land, 2: water, 3: cloud}
    """
    if isinstance(inputs, torch.Tensor):
        mask_invalids = torch.all(inputs == 0, dim=0).cpu()
        output = torch.ones(prediction.shape[-2:], dtype=torch.uint8, device=torch.device("cpu"))
        prediction = prediction.cpu()
    else:
        mask_invalids = np.all(inputs == 0, axis=0)
        output = np.ones(prediction.shape[-2:], dtype=np.uint8)

    cloud_mask = (prediction[0] > th_cloud)
    # erode cloud mask as in create_gt??

    water_mask = (prediction[1] > th_water)
    output[water_mask] = 2
    
    if mask_clouds:
        br = get_brightness(inputs, channels_input=channels_input,
                            collection_name=collection_name) > th_brightness
        if hasattr(br, "cpu"):
            br = br.cpu()
        output[cloud_mask & br] = 3
    else:
        output[cloud_mask] = 3
    
    output[mask_invalids] = 0
    return output

def compute_cloud_coverage(floodmap:gpd.GeoDataFrame, area_imaged:Optional[Union[Polygon, MultiPolygon]]=None) -> float:
    """
    Returns cloud coverage from a vectorized floodmap

    Args:
        floodmap: GeoDataFrame with column 'class' with values 'area_imaged' and 'cloud'
        area_imaged: Polygon that indicates the area imaged. if not provided it will take it from the floodmap

    Returns:
        percentage of the image covered by clouds
    """
    if area_imaged is None:
        area_total = floodmap[floodmap["class"] == "area_imaged"].geometry.area.sum()
    else:
        area_total = area_imaged.area

    area_clouds = floodmap[floodmap["class"] == "cloud"].geometry.area.sum()
    return float(area_clouds / area_total)


def get_area_missing_or_cloud(floodmap:gpd.GeoDataFrame,
                              area_imaged:Union[Polygon,MultiPolygon]) -> Union[Polygon, MultiPolygon]:
    """
    Returns a Polygon with the area of the floodmap that hasn't been imaged (i.e. is out of `floodmap[floodmap['class'] == "area_imaged"]`)
    and that is not covered by clouds.

    Args:
        floodmap: GeoDataFrame with column 'class' with values 'area_imaged' and 'cloud'
        area_imaged: Polygon that indicates the area imaged.

    Returns:
        Polygon with the area of the floodmap that hasn't been imaged and that is not covered by clouds
    """

    area_missing = area_imaged.difference(unary_union(floodmap[(floodmap["class"] == "area_imaged")].geometry))
    clouds = unary_union(floodmap[(floodmap["class"] == "cloud")].geometry)
    area_missing_or_cloud =  clouds.union(area_missing)

    # Remove Lines or Points from missing area
    if area_missing_or_cloud.type == "GeometryCollection":
        area_missing_or_cloud = unary_union(
            [gc for gc in area_missing_or_cloud.geoms if (gc.type == "Polygon") or (gc.type == "MultiPolygon")])

    return area_missing_or_cloud


def get_area_valid(floodmap:gpd.GeoDataFrame) -> Union[Polygon, MultiPolygon]:
    """
    Returns the polygon with the area valid (i.e. area that is imaged and is not cloud)

    Args:
        floodmap: GeoDataFrame with column 'class' with values 'area_imaged' and 'cloud'

    Returns:
        Polygon with the area valid

    """
    area_valid = unary_union(floodmap[(floodmap["class"] == "area_imaged")].geometry)
    clouds = unary_union(floodmap[(floodmap["class"] == "cloud")].geometry)
    return area_valid.difference(clouds)


def get_floodmap_pre(geojsons:List[str], verbose:bool=False) -> gpd.GeoDataFrame:
    """
    From a list of predicted GeoJSONs returns the GeoDataFrame with all water.


    Args:
        geojsons: List[GeoDataFrame] with column 'class' with values {"water", "cloud", "area_imaged", "flood_trace"}
        verbose:

    Returns:
        floodmap with the lowest cloud coverage. Classes: {"water", "cloud", "area_imaged"}

    """

    # fill the clouds in pre with data from other geojsons

    datas = []
    # Compute total area observed
    area_imaged = Polygon() # empty polygon
    crs = None
    for g in geojsons:
        data = utils.read_geojson_from_gcp(g)
        if crs is None:
            crs = data.crs
        elif crs != data.crs:
            data = data.to_crs(crs=crs)
        data = data[["class","geometry"]].copy()

        area_imaged = unary_union(data[data["class"] == "area_imaged"].geometry).union(area_imaged)
        datas.append(data)

    # Compute cloud coverage of all pre-flood data taking into account area observed
    ccs = []
    for data in datas:
        ccs.append(compute_cloud_coverage(data, area_imaged=area_imaged))

    # Sort data based on ccs
    idx_sorted = np.argsort(ccs)
    datas_sorted = [datas[idxdates] for idxdates in idx_sorted]

    return mosaic_floodmaps(datas_sorted, area_imaged, verbose=verbose)


def get_floodmap_post(geojsons:List[str],verbose:bool=False) -> gpd.GeoDataFrame:
    """
    From a list of sorted GeoJSONs returns the GeoDataFrame with all flood water smartly joined

    Args:
        geojsons: List[GeoDataFrame] with column 'class' with values {"water", "cloud", "area_imaged", "flood_trace"}
        verbose:

    Returns:
        floodmap with lowest cloud coverage. Classes: {"water", "cloud", "area_imaged"}

    """

    # fill the clouds in pre with data from other geojsons

    datas = []
    # Compute total area observed
    area_imaged = Polygon() # empty polygon
    crs = None
    for g in geojsons:
        data = utils.read_geojson_from_gcp(g)
        if crs is None:
            crs = data.crs
        elif crs != data.crs:
            data = data.to_crs(crs=crs)
        data = data[["class","geometry"]].copy()

        area_imaged = unary_union(data[data["class"] == "area_imaged"].geometry).union(area_imaged)
        datas.append(data)

    return mosaic_floodmaps(datas, area_imaged, classes_water=["water", "flood_trace"], verbose=verbose)


def mosaic_floodmaps(datas:List[gpd.GeoDataFrame],
                     area_imaged:Union[Polygon, MultiPolygon],
                     classes_water:List[str]=["water"],
                     verbose:bool=False) -> gpd.GeoDataFrame:
    """
    Mosaics the floodmaps iteratively taking into account the valid area of each of them.

    Args:
        datas: List[GeoDataFrame] with column 'class' with values {"water", "cloud", "area_imaged", "flood_trace"}
        area_imaged: Polygon with the total area imaged
        classes_water: which water classes from ["water", "flood_trace"] to include in the output floodmap. For pre-flood
        maps we don't include the flood_trace class.
        verbose:

    Returns:
        a GeoDataFrame with the mosaic of the input dataframes
    """

    # Loop data computing pre-flood water
    warnings.filterwarnings('ignore', 'GeoSeries.isna', UserWarning)
    best_floodmap = datas[0]
    crs = best_floodmap.crs
    area_missing_or_cloud = get_area_missing_or_cloud(best_floodmap, area_imaged=area_imaged)

    # This dataframe will be filled with water polygons
    condition = None
    for c in classes_water:
        if condition is None:
            condition = best_floodmap["class"] == c
        else:
            condition|= (best_floodmap["class"] == c)

    best_floodmap = best_floodmap[condition].copy()
    for idx, data in enumerate(datas[1:]):
        if area_missing_or_cloud.is_empty or not (area_missing_or_cloud.type in ["Polygon", "MultiPolygon", "GeometryCollection"]):
            if verbose:
                print(f"All area is covered in idx {idx+1}. Area missing empty: {area_missing_or_cloud.is_empty} Geom type: {area_missing_or_cloud.type}")
            break

        if data.crs != best_floodmap.crs:
            data = data.to_crs(data.crs, inplace=False)

        # Add water polygons that intersect the missing data
        for c in classes_water:
            water_geoms = data[data["class"] == c].geometry.apply(lambda g: g.intersection(area_missing_or_cloud))

            water_geoms = water_geoms[~water_geoms.isna() & ~water_geoms.is_empty]
            water_geoms = water_geoms.explode(ignore_index=True)
            water_geoms = water_geoms[water_geoms.geometry.type == "Polygon"]

            if water_geoms.shape[0] > 0:
                water_data = gpd.GeoDataFrame(geometry=water_geoms, crs=crs)
                water_data["class"] = c
                best_floodmap = pd.concat([best_floodmap, water_data], ignore_index=True)

        # Update area missing or cloud
        area_missing_or_cloud = get_area_missing_or_cloud(data, area_imaged).intersection(area_missing_or_cloud)

        # Remove points or LineStrings
        if area_missing_or_cloud.type == "GeometryCollection":
            area_missing_or_cloud = unary_union([gc for gc in area_missing_or_cloud.geoms if (gc.type == "Polygon") or (gc.type == "MultiPolygon")])


    # Join adjacent polygons
    best_floodmap = best_floodmap.dissolve(by="class").reset_index()
    # Explode multipoligons to polygons
    best_floodmap = best_floodmap.explode(ignore_index=True)
    stuff_concat = [best_floodmap]

    # Add clouds
    if not area_missing_or_cloud.is_empty and (area_missing_or_cloud.type in ["Polygon", "MultiPolygon", "GeometryCollection"]):
        # If there is something missing must be cloud (because area_imaged is the union of all the area missing in all the pre-floodmaps)
        cloud_data = gpd.GeoDataFrame(geometry=[area_missing_or_cloud], crs=crs)
        cloud_data = cloud_data.explode(ignore_index=True)
        cloud_data["class"] = "cloud"
        stuff_concat.append(cloud_data)

    # Add area_imaged
    area_imaged_data = gpd.GeoDataFrame(geometry=[area_imaged], crs=crs)
    area_imaged_data = area_imaged_data.explode(ignore_index=True)
    area_imaged_data["class"] = "area_imaged"
    stuff_concat.append(area_imaged_data)

    # Filter stuff that are not polygons
    result =  pd.concat(stuff_concat, ignore_index=True)
    assert (result.geometry.type != "MultiPolygon").all(), "Everything should be flattened! found some MultiPolygon"
    assert (result.geometry.type != "GeometryCollection").all(), "Everything should be flattened! found some GeometryCollection"
    # Remove geometries that are not polyongs and exclude polygons with area >= 400m^2
    result = result[(result.geometry.type == "Polygon") & (result.geometry.area >= 20*20)].copy()
    result["geometry"] = result["geometry"].simplify(tolerance=10)
    result = result[~result.geometry.isna() & ~result.geometry.is_empty]

    return result



def compute_pre_post_flood_water(floodmap_post_data:gpd.GeoDataFrame, best_pre_flood_data:gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Computes the difference between water in the pre and post floodmap

    Args:
        floodmap_post_data: Floodmap with classes {'area_imaged' 'water' 'cloud' 'flood_trace'}
        best_pre_flood_data: Floodmap with classes {'area_imaged' 'water' 'cloud' 'flood_trace'}

    Returns:
        Floodmap with classes {'area_imaged'
                               'water-post-flood',
                               'water-pre-flood,
                                'cloud',
                                'flood-trace',
                                'area_imaged-pre-flood',
                                'cloud-pre-flood'}

    """
    warnings.filterwarnings('ignore', 'GeoSeries.isna', UserWarning)

    if floodmap_post_data.crs != best_pre_flood_data.crs:
        best_pre_flood_data = best_pre_flood_data.to_crs(floodmap_post_data.crs)

    area_imaged_post = unary_union(floodmap_post_data[floodmap_post_data["class"] == "area_imaged"].geometry)

    area_missing_pre = get_area_missing_or_cloud(best_pre_flood_data, area_imaged_post)
    pre_flood_water_or_missing_pre = unary_union(best_pre_flood_data[best_pre_flood_data["class"] == "water"].geometry).union(area_missing_pre)

    # pre_flood_cloud = unary_union(best_pre_flood_data[(best_pre_flood_data["class"] == "cloud")].geometry)
    # pre_flood_water_minus_cloud = pre_flood_water.difference(pre_flood_cloud)

    geoms_flood = floodmap_post_data[floodmap_post_data["class"] == "water"].geometry.apply(
        lambda g: g.difference(pre_flood_water_or_missing_pre))
    geoms_flood = geoms_flood[~geoms_flood.isna() & ~geoms_flood.is_empty]
    geoms_flood = geoms_flood.explode(ignore_index=True)
    geoms_flood = geoms_flood[geoms_flood.geometry.type == "Polygon"]

    geoms_trace = floodmap_post_data[(floodmap_post_data["class"] =="flood_trace")].geometry.apply(
        lambda g: g.difference(pre_flood_water_or_missing_pre))

    geoms_trace = geoms_trace[~geoms_trace.isna() & ~geoms_trace.is_empty]
    geoms_trace = geoms_trace.explode(ignore_index=True)
    geoms_trace = geoms_trace[geoms_trace.geometry.type == "Polygon"]

    data_post_flood = gpd.GeoDataFrame(geometry=geoms_flood, crs=floodmap_post_data.crs)
    data_post_flood["class"] = "water-post-flood"
    data_post_flood_trace = gpd.GeoDataFrame(geometry=geoms_trace, crs=floodmap_post_data.crs)
    data_post_flood_trace["class"] = "flood-trace"
    data_post_flood = pd.concat([data_post_flood_trace, data_post_flood], ignore_index=True)

    # Expand multipolygons to single polygons
    data_post_flood = data_post_flood.explode(ignore_index=True)

    # Remove geometries that are not polyongs and exclude polygons with area >= 400m^2
    data_post_flood = data_post_flood[(data_post_flood.geometry.type == "Polygon") & (data_post_flood.geometry.area >= 20 * 20)].copy()

    # simplify polygons
    data_post_flood["geometry"] = data_post_flood["geometry"].simplify(tolerance=10)

    # Copy stuff from pre-flood data
    best_pre_flood_data_propagate = best_pre_flood_data[best_pre_flood_data["class"] != "flood_trace"].copy()
    best_pre_flood_data_propagate.loc[best_pre_flood_data_propagate["class"] == "water", "class"] = "water-pre-flood"
    best_pre_flood_data_propagate.loc[best_pre_flood_data_propagate["class"] == "cloud", "class"] = "cloud-pre-flood"
    best_pre_flood_data_propagate.loc[best_pre_flood_data_propagate["class"] == "area_imaged", "class"] = "area_imaged-pre-flood"

    post_flood_propagate = floodmap_post_data[(floodmap_post_data["class"] == "area_imaged") | (floodmap_post_data["class"] == "cloud")]

    result = pd.concat([best_pre_flood_data_propagate, post_flood_propagate, data_post_flood],
                     ignore_index=True)
    result = result[~result.geometry.isna() & ~result.geometry.is_empty]
    return result


def geometrycollection_to_multipolygon(x:GeometryCollection) -> Union[MultiPolygon, Polygon]:
    if x.type == "GeometryCollection":
        x = unary_union(
            [gc for gc in x.geoms if (gc.type == "Polygon") or (gc.type == "MultiPolygon")])
    return x


def make_valid(df:gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    df["geometry"] = df.geometry.apply(lambda x: geometrycollection_to_multipolygon(validation.make_valid(x)))
    df = df[(~df.geometry.is_empty) & ((df.geometry.type == "Polygon") | (df.geometry.type == "MultiPolygon"))].copy()
    return df.explode(ignore_index=True)


def spatial_aggregation(floodmaps_paths:List[str], dst_crs:str= "EPSG:4326") -> gpd.GeoDataFrame:
    """
    Mosaics all the pre-post floodmaps into a single map reprojecting the polygons to the specified CRS

    Args:
        floodmaps_paths: List of paths to floodmaps as generated with get_floodmap_post and get_floodmaps_pre
        dst_crs:

    Returns:
        geoDataFrame with same classes as floodmaps_paths files

    """
    data_all = None

    for f in tqdm(floodmaps_paths):
        data = utils.read_geojson_from_gcp(f)
        data = data[~data.geometry.isna() & ~data.geometry.is_empty & (data.geometry.area > 10 ** 2)].copy()
        data = data.to_crs(dst_crs)

        is_valid_geoms = data.is_valid
        if not is_valid_geoms.all():
            reasons_invalidity = [f"{validation.explain_validity(g)}\n" for g in data.geometry[~is_valid_geoms]]
            print(f"\tProduct {f} There are {(~is_valid_geoms).sum()} geoms invalid of {is_valid_geoms.shape[0]}\n {reasons_invalidity}")
            data = make_valid(data)

        if data_all is None:
            data_all = data
        else:
            data_all = pd.concat([data_all, data], ignore_index=True)
            # data_all = data_all.dissolve(by="class").reset_index()
            # data_all = data_all.explode(ignore_index=True)
            # data_all = data_all[~data_all.geometry.isna() & ~data_all.geometry.is_empty]

    print(f"\t{len(floodmaps_paths)} Products joined {data_all.shape}")
    # Save as geojson
    data_all = data_all.dissolve(by="class").reset_index()
    print(f"\t{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} Correctly dissolved. New shape: {data_all.shape}")
    data_all = data_all.explode(ignore_index=True)
    print(f"\t{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} Correctly exploded. New shape: {data_all.shape}")

    data_all = data_all[~data_all.geometry.isna() & ~data_all.geometry.is_empty]
    print(f"\t{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} Removed NA or empty polygons. New shape: {data_all.shape}")

    return data_all


def vectorize_jrc_permanent_water_layer(permanent_water:np.ndarray,
                                        crs:str, transform:rasterio.transform.Affine) -> Optional[gpd.GeoDataFrame]:
    """
    Vectorize cloud class

    Args:
        permanent_water: (H, W) array with predictions.
        https://developers.google.com/earth-engine/datasets/catalog/JRC_GSW1_3_YearlyHistory?hl=en
        Values 0 nodata 1 not water, 2 seasonal water, 3: permanent water
        crs:
        transform:

    Returns:
        gpd.GeoDataFrame with column "class" with values "seasonal" and "permanent"
    """

    data_out = []
    for c, class_name in zip([2, 3],["seasonal", "permanent"]):
        geoms_polygons = get_water_polygons(permanent_water == c,
                                            transform=transform)
        if len(geoms_polygons) > 0:
            data_out.append(gpd.GeoDataFrame({"geometry": geoms_polygons,
                                              "class": class_name},
                                             crs=crs))

    if len(data_out) == 1:
        return data_out[0]
    elif len(data_out) > 1:
        return pd.concat(data_out, ignore_index=True)
    return None

def load_vectorized_permanent_water(aoi_folder:str, year:Optional[str] = None) -> Optional[gpd.GeoDataFrame]:
    """
    This function caches the vectorized permanent water product as a GeoJSON if it doesn't exist and return the vectorized
    product as a GeoDataFrame

    Args:
        aoi_folder: in this folder there should exist a subfolder called "PERMANENTWATERJRC" with the tif file
        (JRC water product exported from the GEE)
        https://developers.google.com/earth-engine/datasets/catalog/JRC_GSW1_3_YearlyHistory?hl=en
        year: year to load (must exist in {aoi_folder}/PERMANENTWATERJRC/{year}.tif)

    Returns:
        gpd.GeoDataFrame with column "class" with values "seasonal" and "permanent"

    """
    fs = utils.get_filesystem(aoi_folder)
    if year is None:
        year = "*"
    files_permanent = fs.glob(os.path.join(aoi_folder, "PERMANENTWATERJRC", f"{year}.tif").replace("\\", "/"))
    if len(files_permanent) == 0:
        permanent_water_floodmap = None
    else:
        namefile_permanent = os.path.basename(os.path.splitext(files_permanent[0])[0])
        path_permanent_water_vec = os.path.join(aoi_folder, "PERMANENTWATERJRC_vec",
                                                namefile_permanent + ".geojson").replace("\\", "/")
        if not fs.exists(path_permanent_water_vec):
            with rasterio.open(f"gs://{files_permanent[0]}") as rst:
                permanent_raster = rst.read(1)
                crs = rst.crs
                transform = rst.transform

            permanent_water_floodmap = vectorize_jrc_permanent_water_layer(permanent_raster, crs=crs,
                                                                           transform=transform)
            if permanent_water_floodmap is not None:
                utils.write_geojson_to_gcp(path_permanent_water_vec, permanent_water_floodmap)
        else:
            permanent_water_floodmap = utils.read_geojson_from_gcp(path_permanent_water_vec)

    return permanent_water_floodmap

def add_permanent_water_to_floodmap(jrc_vectorized_map:gpd.GeoDataFrame, floodmap:gpd.GeoDataFrame,
                                    water_class:Optional[str]=None) -> gpd.GeoDataFrame:
    """
    Adds the "permanent" polygons of jrc_vectorized_map to floodmap

    Args:
        jrc_vectorized_map:
        floodmap:
        water_class:

    Returns:
        floodmap with permanent water polygons
    """
    classes = floodmap["class"].unique()
    if water_class is None:
        if "water" in classes:
            water_class = "water"
        elif "water-pre-flood" in classes:
            water_class = "water-pre-flood"
        else:
            raise AttributeError(f"water_class not provided and we couldn't guess it")

    jrc_vectorized_map_copy = jrc_vectorized_map[jrc_vectorized_map["class"] == "permanent"].copy()
    jrc_vectorized_map_copy = jrc_vectorized_map_copy[["geometry", "class"]]
    if jrc_vectorized_map_copy.shape[0] == 0:
        return floodmap

    jrc_vectorized_map_copy["class"] = water_class

    floodmap = pd.concat([floodmap, jrc_vectorized_map_copy], ignore_index=True)
    floodmap = floodmap.dissolve(by="class").reset_index()
    floodmap = floodmap.explode(ignore_index=True)

    # fix water in flood_trace or flood-trace classes
    if "flood_trace" in classes:
        class_flood_trace = "flood_trace"
    elif "flood-trace" in classes:
        class_flood_trace = "flood-trace"
    else:
        return floodmap

    water_polygon = unary_union(floodmap[floodmap["class"] == water_class].geometry)
    geoms_trace = floodmap[(floodmap["class"] == class_flood_trace)].geometry.apply(
        lambda g: g.difference(water_polygon))

    geoms_trace = geoms_trace[~geoms_trace.isna() & ~geoms_trace.is_empty]
    geoms_trace = geoms_trace.explode(ignore_index=True)
    geoms_trace = geoms_trace[geoms_trace.geometry.type == "Polygon"]

    # Add back to floodmap
    floodmap = floodmap[floodmap["class"] != class_flood_trace].reset_index()

    floodmap_trace = gpd.GeoDataFrame(geometry=geoms_trace, crs=floodmap.crs)
    floodmap_trace["class"] = class_flood_trace

    floodmap = pd.concat([floodmap, floodmap_trace], ignore_index=True)
        
    return floodmap




from ml4floods.data.worldfloods.dataset import WorldFloodsDatasetTiled
from ml4floods.preprocess.tiling import WindowSlices
from typing import List, Callable
from tqdm import trange
import numpy as np

def _filter_windows(fun_frac_clouds_invalids:Callable,
                    dataset:WorldFloodsDatasetTiled, threshold_clouds:float=.5) ->List[WindowSlices]:
    """ filter windows from the dataset with more tham threshold_clouds * 100 of clouds or invalids """

    valid_slices = []
    for idx in trange(len(dataset), desc="Filtering invalid and cloudy windows"):
        label = dataset.get_label(idx)
        frac_invalids = fun_frac_clouds_invalids(label)
        if frac_invalids < threshold_clouds:
            valid_slices.append(dataset.list_of_windows[idx])

    return valid_slices


def filter_windows_v1(dataset: WorldFloodsDatasetTiled, threshold_clouds: float = .5) -> List[WindowSlices]:
    """ filter windows from the dataset with more that threshold_clouds * 100 of clouds or invalids """

    return _filter_windows(lambda label: ((label == 0) | (label == 3)).sum()/np.prod(label.shape),
                           dataset, threshold_clouds=threshold_clouds)


def filter_windows_v2(dataset: WorldFloodsDatasetTiled, threshold_clouds: float = .5) -> List[WindowSlices]:
    """ filter windows from the dataset with more that threshold_clouds * 100 of clouds or invalids """

    # Assumes first channel is cloud second channel is water
    return _filter_windows(lambda label: (label[1] == 0).sum() / np.prod(label.shape[1:]),
                           dataset, threshold_clouds=threshold_clouds)






import torch
import torch.nn as nn

from ml4floods.models.architectures import layer_factory


class UNet(nn.Module):
    def __init__(self, n_channels, n_class):
        super().__init__()

        self.dconv_down1 = layer_factory.double_conv(n_channels, 64)
        self.dconv_down2 = layer_factory.double_conv(64, 128)
        self.dconv_down3 = layer_factory.double_conv(128, 256)
        self.dconv_down4 = layer_factory.double_conv(256, 512)

        self.maxpool = nn.MaxPool2d(2)

        self.dconv_up3 = layer_factory.double_conv(256 + 512, 256)
        self.dconv_up2 = layer_factory.double_conv(128 + 256, 128)
        self.dconv_up1 = layer_factory.double_conv(128 + 64, 64)

        self.conv_last = nn.Conv2d(64, n_class, 1)

    def forward(self, x):
        conv1 = self.dconv_down1(x)
        x = self.maxpool(conv1)

        conv2 = self.dconv_down2(x)
        x = self.maxpool(conv2)

        conv3 = self.dconv_down3(x)
        x = self.maxpool(conv3)

        x = self.dconv_down4(x)

        x = nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=True)
        x = torch.cat([x, conv3], dim=1)

        x = self.dconv_up3(x)
        x = nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=True)
        x = torch.cat([x, conv2], dim=1)

        x = self.dconv_up2(x)
        x = nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=True)
        x = torch.cat([x, conv1], dim=1)

        x = self.dconv_up1(x)

        out = self.conv_last(x)

        return out

    
class UNet_dropout(nn.Module):
    def __init__(self, n_channels, n_class):
        super().__init__()

        self.dconv_down1 = layer_factory.double_conv(n_channels, 64)
        self.dconv_down2 = layer_factory.double_conv(64, 128)
        self.dconv_down3 = layer_factory.double_conv(128, 256)
        self.dconv_down4 = layer_factory.double_conv(256, 512)

        self.maxpool = nn.MaxPool2d(2)

        self.dconv_up3 = layer_factory.double_conv(256 + 512, 256)
        self.dconv_up2 = layer_factory.double_conv(128 + 256, 128)
        self.dconv_up1 = layer_factory.double_conv(128 + 64, 64)
        
        self.dropout = nn.Dropout2d()

        self.conv_last = nn.Conv2d(64, n_class, 1)

    def forward(self, x):
        conv1 = self.dropout(self.dconv_down1(x))
        x = self.maxpool(conv1)

        conv2 = self.dropout(self.dconv_down2(x))
        x = self.maxpool(conv2)

        conv3 = self.dropout(self.dconv_down3(x))
        x = self.maxpool(conv3)

        x = self.dropout(self.dconv_down4(x))

        x = nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=True)
        x = torch.cat([x, conv3], dim=1)

        x = self.dropout(self.dconv_up3(x))
        x = nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=True)
        x = torch.cat([x, conv2], dim=1)

        x = self.dropout(self.dconv_up2(x))
        x = nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=True)
        x = torch.cat([x, conv1], dim=1)

        x = self.dropout(self.dconv_up1(x))

        out = self.conv_last(x)

        return out



import pickle

from ml4floods.data.index.map_data import MapData


class MapDataFactory:
    def __init__(self, index_path: str):
        """
        Initialize a MapDataFactory from a geographic index that is capable of producing MapData objects given coordinate rectangles.
        """
        with open(index_path, 'rb') as f:
            self.index_list = pickle.load(f)

    def create_map_data(self, min_lat: int, min_lon: int, max_lat: int, max_lon: int):
        """
        Create a MapData object given a coordinate rectangle. Note that max_lat and max_lon are exclusive integral upper bounds.
        """
        return MapData(min_lat, min_lon, max_lat, max_lon, self.index_list)



import logging
import pickle
from dataclasses import dataclass
from typing import List

LOGGER = logging.getLogger(__name__)


class GeographicIndex:

    @staticmethod
    def _initialize_index() -> List[str]:
        """
        This function initializes a geographical index.
        
        We create a list of lists of strings. Each string corresponds
        to a piece of data. The list is indexed by a latitude - 90 and
        longitude - 180 using the formula (lat-90)*360 + (lon-180). 
        Transforming the 2D array to 1D using lat long.

        Args:
          None

        Returns:
          A list of lists of strings.
        """

        index_list = []
        for i in range(180 * 360):
            index_list.append([])
        return index_list

    def append_at_coords(self, lat: int, lon: int, records: list):
        """
        Append a list of records to the geographical index for a specific integral lat/lon location
        """
        flattened_idx = (lat + 90) * 360 + lon + 180
        LOGGER.debug(f"Append records to index: lat = {lat} lon = {lon} flat = {flattened_idx}")
        for rec in records:
            self.index_list[flattened_idx].append(rec)

    def save_index(self, path: str):
        with open(path, 'wb') as f:
            pickle.dump(self.index_list, f)

    def load_index(self, path: str):
        with open(path, 'rb') as f:
            self.index_list = pickle.load(f)

    def __init__(self):
        self.index_list = GeographicIndex._initialize_index()




from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import rasterio
import rasterio.windows
from torch.utils.data import Dataset
from ml4floods.preprocess.tiling import WindowSize
from ml4floods.preprocess.utils import get_list_of_window_slices


class RasterioImageDataset(Dataset):
    """A prepackaged WorldFloods PyTorch Dataset
    This initializes the dataset given a set a set of image files with a
    subdirectory for the training and testing data ("image_prefix" and "gt_prefix").

    Args:
        image_files (List[str]): the image files to be loaded into the
            dataset
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        transforms (Callable): the transformations used within the
            training data module

    Attributes:
        image_files (List[str]): the image files to be loaded into the
            dataset
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        transforms (Callable): the transformations used within the
            training data module
    """

    def __init__(
        self,
        image_files: List[str],
        image_prefix: str = "/image_files/",
        gt_prefix: str = "/gt_files/",
        transforms: Optional[List[Callable]] = None,
    ) -> None:

        self.image_files = image_files
        self.image_prefix = image_prefix
        self.gt_prefix = gt_prefix
        self.transforms = transforms

        # sort to make sure that the order is deterministic
        # (order of the flow of data points to the ML model)
        # TODO: Do this for the list of filepaths at the end as well
        self.image_files.sort()

    def __getitem__(self, idx: int) -> Dict:
        """Index to select an image

        Args:
            idx (int): index

        Returns:
            a dictionary with the image and mask keys
            {"image", "mask"}
        """

        # get filenames
        image_name = self.image_files[idx]

        y_name = image_name.replace(self.image_prefix, self.gt_prefix, 1)

        # Open Image File
        with rasterio.open(image_name) as f:
            image_tif = f.read()

        # Open Ground Truth File
        with rasterio.open(y_name) as f:
            mask_tif = f.read()

        # get rid of nan, convert to float
        image = np.nan_to_num(image_tif).astype(np.float32)

        # TODO: Need to check why the 0th index.
        mask = np.nan_to_num(mask_tif)

        # Apply transformation
        if self.transforms is not None:
            data = self.transforms(image=image, mask=mask)
        else:
            data = {"image": image, "mask": mask}
        # return x, y
        return data

    def __len__(self) -> int:
        return len(self.image_files)


class RasterioImageTilingDataset(Dataset):
    """A prepackaged WorldFloods PyTorch Dataset
    This initializes the dataset given a set a set of image files with a
    subdirectory for the training and testing data ("image_prefix" and "gt_prefix").
    This also does the tiling under the hood given the windowsize.

    Args:
        image_files (List[str]): the image files to be loaded into the
            dataset
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        window_size (Tuple[int,int]): the window sizes (height, width) to be
            used for the tiling
        transforms (Callable): the transformations used within the
            training data module

    Attributes:
        image_files (List[str]): the image files to be loaded into the
            dataset
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        window_size (namedtuple): a tuple with the height and width
            arguments
        transforms (Callable): the transformations used within the
            training data module
        accumulated_list_of_windows_test (List[namedtuple]): a list of
            namedtuples each consisting of a filename and a rasterio.window
    """

    def __init__(
        self,
        image_files: List[str],
        image_prefix: str = "/image_files/",
        gt_prefix: str = "/gt_files/",
        window_size: Tuple[int, int] = (64, 64),
        transforms: Optional[List[Callable]] = None,
    ) -> None:

        self.image_files = image_files
        self.image_prefix = image_prefix
        self.gt_prefix = gt_prefix
        self.transforms = transforms
        self.window_size = WindowSize(height=window_size[0], width=window_size[1])

        # sort to make sure that the order is deterministic
        # (order of the flow of data points to the ML model)
        # TODO: Do this for the list of filepaths at the end as well
        self.image_files.sort()
        # get the image slices
        self.accumulated_list_of_windows_test = get_list_of_window_slices(
            self.image_files, window_size=self.window_size
        )

    def __getitem__(self, idx: int) -> Dict:
        """Index to select an image tile

        Args:
            idx (int): index

        Returns:
            a dictionary with the keys to the image tiles and mask tiles
            {"image", "mask"}
        """
        # get filenames from named tuple
        sub_window = self.accumulated_list_of_windows_test[idx]

        # get filename
        image_name = sub_window.file_name

        # replace string for image_prefix
        image_name = image_name.replace(self.gt_prefix, self.image_prefix, 1)

        # replace string for gt_prefix
        y_name = image_name.replace(self.image_prefix, self.gt_prefix, 1)

        # Open Image File
        with rasterio.open(image_name) as f:
            image_tif = f.read(window=sub_window.window, boundless=True, fill_value=0)

        # Open Ground Truth File
        with rasterio.open(y_name) as f:
            mask_tif = f.read(window=sub_window.window, boundless=True, fill_value=0)

        # get rid of nan, convert to float
        image = np.nan_to_num(image_tif).astype(np.float32)

        # TODO: Need to check why the 0th index.
        mask = np.nan_to_num(mask_tif)

        # Apply transformation
        if self.transforms is not None:
            data = self.transforms(image=image, mask=mask)
        else:
            data = {"image": image, "mask": mask}
        # return x, y
        return data

    def __len__(self) -> int:
        return len(self.accumulated_list_of_windows_test)


from ml4floods.models.config_setup import get_default_config, get_filesystem
from ml4floods.models.model_setup import get_model
from ml4floods.models.model_setup import get_model_inference_function
from ml4floods.models.model_setup import get_channel_configuration_bands
from ml4floods.models.utils.configuration import AttrDict
from ml4floods.data.worldfloods import dataset
from ml4floods.data import create_gt
from ml4floods.models import postprocess
import torch
import rasterio
from ml4floods.data import save_cog, utils
import numpy as np
import os
from datetime import datetime
import geopandas as gpd
import pandas as pd
import warnings
import sys
import traceback
from ml4floods.models.postprocess import get_pred_mask_v2
from typing import Tuple, Callable, Any, Optional
from ml4floods.data.worldfloods.configs import BANDS_S2, BANDS_L8
from skimage.morphology import binary_dilation, disk


def load_inference_function(model_path: str, device_name: str, max_tile_size: int = 1024,
                            th_water: float = .5,
                            th_brightness: float = create_gt.BRIGHTNESS_THRESHOLD,
                            collection_name:str="S2",
                            distinguish_flood_traces:bool=False) -> Tuple[
    Callable[[torch.Tensor], Tuple[torch.Tensor,torch.Tensor]], AttrDict]:
    if model_path.endswith("/"):
        experiment_name = os.path.basename(model_path[:-1])
        model_folder = os.path.dirname(model_path[:-1])
    else:
        experiment_name = os.path.basename(model_path)
        model_folder = os.path.dirname(model_path[:-1])

    config_fp = os.path.join(model_path, "config.json").replace("\\", "/")
    config = get_default_config(config_fp)

    # The max_tile_size param controls the max size of patches that are fed to the NN. If you're in a memory constrained environment set this value to 128
    config["model_params"]["max_tile_size"] = max_tile_size

    config["model_params"]['model_folder'] = model_folder
    config["model_params"]['test'] = True
    model = get_model(config.model_params, experiment_name)
    model.to(device_name)
    inference_function = get_model_inference_function(model, config, apply_normalization=True,
                                                      activation=None)

    if config.model_params.get("model_version", "v1") == "v2":

        channels = get_channel_configuration_bands(config.data_params.channel_configuration,
                                                   collection_name=collection_name)
        if distinguish_flood_traces:
            if collection_name == "S2":
                band_names_current_image = [BANDS_S2[iband] for iband in channels]
                mndwi_indexes_current_image = [band_names_current_image.index(b) for b in ["B3", "B11"]]
            elif collection_name == "Landsat":
                band_names_current_image = [BANDS_L8[iband] for iband in channels]
                # TODO ->  if not all(b in band_names_current_image for b in ["B3","B6"])
                mndwi_indexes_current_image = [band_names_current_image.index(b) for b in ["B3", "B6"]]

        # Add post-processing of binary mask
        def predict(s2l89tensor: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
            """
            Args:
                s2l89tensor: (C, H, W) tensor

            Returns:
                (H, W) mask with interpretation {0: invalids, 1: land, 2: water, 3: think cloud, 4: flood trace}
            """
            with torch.no_grad():
                pred = inference_function(s2l89tensor.unsqueeze(0))[0]  # (2, H, W)
                land_water_cloud =  get_pred_mask_v2(s2l89tensor, pred, channels_input=channels,
                                                     th_water=th_water, th_brightness=th_brightness,
                                                     collection_name=collection_name)

                # Set invalids in continuous pred to -1
                invalids = land_water_cloud == 0
                pred[0][invalids] = -1
                pred[1][invalids] = -1

                if distinguish_flood_traces:
                    s2l89mndwibands = s2l89tensor[mndwi_indexes_current_image, ...].float()

                    # Green − SWIR1)/(Green + SWIR1)
                    mndwi = (s2l89mndwibands[0] - s2l89mndwibands[1]) / (s2l89mndwibands[0] + s2l89mndwibands[1] + 1e-6)

                    land_water_cloud[(land_water_cloud == 2) & (mndwi < 0)] = 4

            return land_water_cloud, pred

    else:
        def predict(s2l89tensor: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
            """
            Args:
                s2l89tensor: (C, H, W) tensor

            Returns:
                (H, W) mask with interpretation {0: invalids, 1: land, 2: water, 3: cloud, 4: flood trace}
            """
            with torch.no_grad():
                pred = inference_function(s2l89tensor.unsqueeze(0))[0]  # (3, H, W)
                invalids = torch.all(s2l89tensor == 0, dim=0)  # (H, W)
                land_water_cloud = torch.argmax(pred, dim=0).type(torch.uint8) + 1  # (H, W)
                land_water_cloud[invalids] = 0

                # Set invalids in continuous pred to -1
                pred[0][invalids] = -1
                pred[1][invalids] = -1
                pred[2][invalids] = -1

                if distinguish_flood_traces:
                    s2l89mndwibands = s2l89tensor[mndwi_indexes_current_image, ...].float()
                    # Green − SWIR1)/(Green + SWIR1)
                    mndwi = (s2l89mndwibands[0] - s2l89mndwibands[1]) / (s2l89mndwibands[0] + s2l89mndwibands[1] + 1e-6)
                    land_water_cloud[(land_water_cloud == 2) & (mndwi < 0)] = 4

            return land_water_cloud, pred

    return predict, config


@torch.no_grad()
def main(model_path: str, s2folder_file: str, device_name: str,
         output_folder: Optional[str]=None, max_tile_size: int = 1_024,
         th_brightness: float = create_gt.BRIGHTNESS_THRESHOLD, th_water: float = .5, overwrite: bool = False,
         collection_name: str = "S2",distinguish_flood_traces:bool=False):

    # This takes into account that this could be run on windows
    s2folder_file = s2folder_file.replace("\\", "/")
    model_path = model_path.replace("\\", "/")

    if model_path.endswith("/"):
        experiment_name = os.path.basename(model_path[:-1])
    else:
        experiment_name = os.path.basename(model_path)

    inference_function, config = load_inference_function(model_path, device_name, max_tile_size=max_tile_size,
                                                         th_water=th_water, th_brightness=th_brightness,
                                                         collection_name=collection_name,
                                                         distinguish_flood_traces=distinguish_flood_traces)

    # Get S2 files to run predictions
    fs = get_filesystem(s2folder_file)
    if s2folder_file.endswith(".tif"):
        s2files = [s2folder_file]
    else:
        if not s2folder_file.endswith("/"):
            s2folder_file += "/"
        s2files = fs.glob(f"{s2folder_file}*.tif")
        if s2folder_file.startswith("gs://"):
            s2files = [f"gs://{s2}" for s2 in s2files]

        assert len(s2files) > 0, f"No Tiff files found in {s2folder_file}*.tif"

    files_with_errors = []

    if output_folder:
        fs_dest = get_filesystem(output_folder)
        if output_folder.endswith("/"):
            output_folder_iter_vec = output_folder[:-1] + "_vec"
            output_folder_iter_cont = output_folder[:-1] + "_cont"
        else:
            output_folder_iter_vec = output_folder + "_vec"
            output_folder_iter_cont = output_folder + "_cont"
    else:
        fs_dest = fs

    for total, filename in enumerate(s2files):
        # Compute folder name to save the predictions if not provided
        if not output_folder:
            base_output_folder = os.path.dirname(os.path.dirname(filename))
            output_folder_iter = os.path.join(base_output_folder, experiment_name, collection_name).replace("\\", "/")
            output_folder_iter_vec = os.path.join(base_output_folder, experiment_name + "_vec", collection_name).replace("\\", "/")
            output_folder_iter_cont = os.path.join(base_output_folder, experiment_name + "_cont",
                                                  collection_name).replace("\\", "/")
        else:
            output_folder_iter = output_folder


        filename_save = os.path.join(output_folder_iter, os.path.basename(filename))
        filename_save_cont = os.path.join(output_folder_iter_cont, os.path.basename(filename))
        filename_save_vect = os.path.join(output_folder_iter_vec,
                                          f"{os.path.splitext(os.path.basename(filename))[0]}.geojson")

        if not overwrite and all(fs_dest.exists(f) for f in [filename_save, filename_save_cont, filename_save_vect]):
            continue

        print(
            f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} ({total + 1}/{len(s2files)}) Processing {collection_name} {filename}")
        if not output_folder:
            print(f"Predictions will be saved in folder: {output_folder_iter}")
        try:
            channels = get_channel_configuration_bands(config.data_params.channel_configuration,
                                                       collection_name=collection_name,as_string=True)
            torch_inputs, transform = dataset.load_input(filename,
                                                         window=None, channels=channels)

            with rasterio.open(filename) as src:
                crs = src.crs

            prediction, pred_cont = inference_function(torch_inputs)
            prediction = prediction.cpu().numpy()
            pred_cont = pred_cont.cpu().numpy()

            # Save data as vectors
            data_out = vectorize_outputv1(prediction, crs, transform)
            if data_out is not None:
                if not filename_save_vect.startswith("gs"):
                    fs_dest.makedirs(os.path.dirname(filename_save_vect), exist_ok=True)
                utils.write_geojson_to_gcp(filename_save_vect, data_out)

            # Save data as COG GeoTIFF
            profile = {"crs": crs,
                       "transform": transform,
                       "compression": "lzw",
                       "RESAMPLING": "NEAREST",
                       "nodata": 0}

            if not filename_save.startswith("gs"):
                fs_dest.makedirs(os.path.dirname(filename_save), exist_ok=True)

            save_cog.save_cog(prediction[np.newaxis], filename_save, profile=profile.copy(),
                              descriptions=["invalid/land/water/cloud/trace"],
                              tags={"invalid":0, "land":1, "water":2, "cloud":3 , "trace":4, "model": experiment_name})

            if not filename_save_cont.startswith("gs"):
                fs_dest.makedirs(os.path.dirname(filename_save_cont), exist_ok=True)

            if pred_cont.shape[0] == 2:
                descriptions = ["clear/cloud", "land/water"]
            else:
                descriptions = ["prob_clear","prob_water", "prob_cloud"]

            profile["nodata"] = -1
            save_cog.save_cog(pred_cont, filename_save_cont, profile=profile.copy(),
                              descriptions=descriptions,
                              tags={"model": experiment_name})

        except Exception:
            warnings.warn(f"Failed")
            traceback.print_exc(file=sys.stdout)
            files_with_errors.append(filename)

    if len(files_with_errors) > 0:
        print(f"Files with errors:\n {files_with_errors}")


def vectorize_outputv1(prediction: np.ndarray, crs: Any, transform: rasterio.Affine,
                       border:int=2) -> Optional[gpd.GeoDataFrame]:
    """

    Args:
        prediction: (H, W) array with 4 posible values  0: "invalid", 2: "water", 3: "cloud", 4: "flood_trace"
        crs:
        transform:
        border:

    Returns:

    """
    data_out = []
    start = 0
    class_name = {0: "area_imaged", 2: "water", 3: "cloud", 4: "flood_trace"}
    # Dilate invalid mask
    invalid_mask = binary_dilation(prediction == 0, disk(3)).astype(np.bool)

    # Set borders to zero to avoid border effects when vectorizing
    prediction[:border,:] = 0
    prediction[:, :border] = 0
    prediction[-border:, :] = 0
    prediction[:, -border:] = 0

    prediction[invalid_mask] = 0
    for c, cn in class_name.items():
        if c == 0:
            # To remove stripes in area imaged
            mask = prediction != c
        else:
            mask = prediction == c

        geoms_polygons = postprocess.get_water_polygons(mask, transform=transform)
        if len(geoms_polygons) > 0:
            data_out.append(gpd.GeoDataFrame({"geometry": geoms_polygons,
                                              "id": np.arange(start, start + len(geoms_polygons)),
                                              "class": cn},
                                             crs=crs))
        start += len(geoms_polygons)

    if len(data_out) == 1:
        return data_out[0]
    elif len(data_out) > 1:
        return pd.concat(data_out, ignore_index=True)

    return None



from typing import List, Dict, Any, Callable, Optional
import numpy as np
import torch
import torch.utils.data
from tqdm import tqdm
import matplotlib.pyplot as plt
import json
import seaborn as sns
import pandas as pd
from collections import OrderedDict



@torch.no_grad()
def compute_confusions(ground_truth_outputs: torch.Tensor,
                       test_outputs_categorical: torch.Tensor, num_class: int,
                       remove_class_zero:bool=False) -> torch.Tensor:
    """
    Compute the confusion matrix for a pair of ground truth and predictions. Returns one confusion matrix for each
    element in the batch

    Args:
        ground_truth_outputs: (B,H,W) tensor with discrete values in [0, num_class] if remove_class_zero else [0,num_class-1]
        test_outputs_categorical: (B,H,W) tensor with discrete values in [0,num_class-1] (expected output of torch.argmax())
        num_class: Number of classes. Must be equal to number of channels of test_outputs_categorical
        remove_class_zero: if true the value zero in ground_truth_outputs is considered a masked value;
        thus removed from the final prediction.

    Returns:
        (B,num_class,num_class) torch.Tensor with a confusion matrix for each image in the batch
        cm[a, b, c] is the number of elements predicted `b` that belonged to class `c` in the image `a` of the batch

    """
    ground_truth = ground_truth_outputs.clone()

    if remove_class_zero:
        # Save invalids to discount
        invalids = ground_truth == 0  # (batch_size, H, W) gpu
        ground_truth[invalids] = 1
        ground_truth -= 1
    
        # Set invalids in pred to zero
        test_outputs_categorical = test_outputs_categorical.clone()
        test_outputs_categorical[invalids] = 0  # (batch_size, H, W)

    confusions_batch = torch.zeros(size=(ground_truth.shape[0], num_class, num_class),
                                   dtype=torch.long)
    for c in range(num_class):
        gtc = (ground_truth == c)
        for c1 in range(num_class):
            pred_c1 = (test_outputs_categorical == c1)
            confusions_batch[:, c1, c] = (pred_c1 & gtc).sum(dim=(1, 2))
    
    if remove_class_zero:
        inv_substract = torch.sum(invalids, dim=(1, 2)).to(confusions_batch.device)
        confusions_batch[:, 0, 0] -= inv_substract
    
    return confusions_batch


def cm_analysis(cm: np.ndarray, labels: List[int], figsize=(10, 10)):
    """
    Generate matrix plot of confusion matrix with pretty annotations.
    The plot image is saved to disk.

    Args:
      cm:  Array with shape (batch_size, len(labels), len(labels))
      labels:    string array, name the order of class labels in the confusion matrix.
                 use `clf.classes_` if using scikit-learn models.
                 with shape (nclass,).
      figsize:   the size of the figure plotted.
    """
    
    cm_sum = np.sum(cm, axis=1, keepdims=True)
    cm_perc = cm / cm_sum.astype(float) * 100
    annot = np.empty_like(cm).astype(str)
    nrows, ncols = cm.shape
    for i in range(nrows):
        for j in range(ncols):
            c = cm[i, j]
            p = cm_perc[i, j]
            if i == j:
                s = cm_sum[i]
                annot[i, j] = '%.1f%%\n%d/%d' % (p, c, s)
            elif c == 0:
                annot[i, j] = ''
            else:
                annot[i, j] = '%.1f%%\n%d' % (p, c)    
    cm = pd.DataFrame(cm_perc, index=labels, columns=labels)
    cm.index.name = 'Predicted'
    cm.columns.name = 'Actual'
    fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(cm, annot=annot, fmt='', ax=ax)
    plt.show()

def binary_accuracy(cm_agg)->float:
    tp = cm_agg[1, 1]
    tn = cm_agg[0, 0]

    return (tp+tn) / (cm_agg.sum() + 1e-6)

def binary_precision(cm_agg) -> float:
    tp = cm_agg[1, 1]
    fp = cm_agg[1, 0]
    return tp / (tp + fp + 1e-6)

def binary_recall(cm_agg) -> float:
    tp = cm_agg[1, 1]
    fn = cm_agg[0, 1]
    return tp / (tp + fn + 1e-6)


def calculate_iou(confusions: torch.Tensor, labels: List[str]) ->Dict[str, float]:
    """
    Caculate IoU for a list of confusion matrices 
    
    Args:
        confusions: List with shape (batch_size, len(labels), len(labels))
        labels: List of class names
        
        returns: dictionary of class names and iou scores for that class (summed across whole matrix list)
    """
    confusions = np.array(confusions)
    conf_matrix = np.sum(confusions, axis=0)
    true_positive = np.diag(conf_matrix) + 1e-6
    false_negative = np.sum(conf_matrix, 0) - true_positive
    false_positive = np.sum(conf_matrix, 1) - true_positive
    iou = true_positive / (true_positive + false_positive + false_negative)
    
    iou_dict = {}
    for i, l in enumerate(labels):
        iou_dict[l] = iou[i]
    return iou_dict


def calculate_recall(confusions: torch.Tensor, labels: List[str]) ->Dict[str, float]:
    confusions = np.array(confusions)
    conf_matrix = np.sum(confusions, axis=0)
    true_positive = np.diag(conf_matrix) + 1e-6
    false_negative = np.sum(conf_matrix, 0) - true_positive
    recall = true_positive / (true_positive + false_negative  + 1e-6)

    recall_dict = {}
    for i, l in enumerate(labels):
        recall_dict[l] = recall[i]
    return recall_dict


def calculate_precision(confusions: torch.Tensor, labels: List[str]) ->Dict[str, float]:
    confusions = np.array(confusions)
    conf_matrix = np.sum(confusions, axis=0)
    true_positive = np.diag(conf_matrix) + 1e-6
    false_positive = np.sum(conf_matrix, 1) - true_positive
    precision = true_positive / (true_positive + false_positive  + 1e-6)

    precision_dict = {}
    for i, l in enumerate(labels):
        precision_dict[l] = precision[i]
    return precision_dict


def plot_metrics(metrics_dict, label_names, thresholds_water = None):
    """
    Plot confusion matrix, precision/recall curve,  tp_rate/fp_rate curve, class IoUs
    
    Args:
        metrics_dict: metrics dictionary as output by 'compute_metrics'
        labels_names: list of class label names
        
        returns: None
    """
    confusions = np.array(metrics_dict['confusions']).transpose(0,2,1)
    confusions_thresh = metrics_dict['confusions_thresholded']
    
    cm_analysis(np.sum(np.array(confusions), axis=0), labels=label_names)

    total_conf_thresh = np.sum(np.array(confusions_thresh), axis=1)
    tp_rates = []
    fp_rates = []

    precisions = []
    recalls = []

    for i in range(len(total_conf_thresh)):
        cur_conf = total_conf_thresh[i]

        tp = cur_conf[1,1]
        tn = cur_conf[0,0]
        fp = cur_conf[1,0]
        fn = cur_conf[0,1]

        tp_rate = tp / (tp+fn)
        fp_rate = fp / (fp+tn)

        tp_rates.append(tp_rate)
        fp_rates.append(fp_rate)

        precision = tp / (tp+fp)
        recall = tp / (tp+fn)

        precisions.append(precision)
        recalls.append(recall)

    fig, ax = plt.subplots(1, 2, figsize=(16, 9))
    sns.set_style('darkgrid')
    plot_df = pd.DataFrame({"Precision": precisions, "Recall": recalls, "FP Rate": fp_rates, "TP Rate": tp_rates})
    sns.lineplot(data=plot_df, x="Recall", y="Precision", ax=ax[0])
    if thresholds_water is not None:
        for i, (rec, prec)  in enumerate(zip(plot_df['Recall'],plot_df['Precision'])):
            ax[0].text(rec, prec, '%10.3f' % thresholds_water[i])
        for i, (fps, tps)  in enumerate(zip(plot_df['FP Rate'],plot_df['TP Rate'])):
            ax[1].text(fps, tps, '%10.3f' % thresholds_water[i])

    sns.lineplot(data=plot_df, x="FP Rate", y="TP Rate", ax=ax[1])

    plt.show()
    
    print("Per Class IOU", json.dumps(metrics_dict['iou'], indent=4, sort_keys=True))
  
def convert_targets_to_v1(target: torch.Tensor) -> torch.Tensor:
    
    clear_clouds = target[:,0,:,:]
    land_water = target[:,1,:,:]
    
    v1gt = land_water.clone() # {0: invalid, 1: land, 2: water}
    v1gt[clear_clouds == 2] = 3

    return v1gt.unsqueeze(dim = 1)
      
def compute_metrics(dataloader:torch.utils.data.dataloader.DataLoader,
                    pred_fun: Callable, thresholds_water=np.arange(0,1,.05),
                    threshold:float=.5,
                    plot=False, mask_clouds:bool=False, convert_targets:bool = True) -> Dict:
    """
    Run inference on a dataloader and compute metrics for that data
    
    Args:
        dataloader: pytorch Dataloader for test set
        pred_fun: function to perform inference using a model
        thresholds_water: list of threshold for precision/recall curves
        threshold: threshold to compute the confusion matrix
        plot: flag for calling plot method with metrics
        mask_clouds: if True this will compute the confusion matrices for "land" and "water" classes only masking
            the cloudy pixels. If mask_clouds the output confusion matrices will be (n_samples, 2, 2) otherwise they will
            be (n_samples, 3, 3)
        convert_targets: if True converts targets from v2 to v1
        
        returns: dictionary of metrics
    """
    confusions = []
    
    # Sort thresholds from high to low
    thresholds_water = np.sort(thresholds_water)
    thresholds_water = thresholds_water[-1::-1]
    confusions_thresh = []

    # This is constant: we're using this class convention to compute the PR curve
    if mask_clouds:
        num_class, label_names = 2, ["land", "water"]
    else:
        num_class, label_names = 3, ["land", "water", "cloud"]

    for i, batch in tqdm(enumerate(dataloader), total=int(len(dataloader.dataset)/dataloader.batch_size)):
        test_inputs, ground_truth = batch["image"], batch["mask"]

        test_outputs = pred_fun(test_inputs)
        
        if convert_targets:
            ground_truth = convert_targets_to_v1(ground_truth)
            
        if mask_clouds:
            # assert test_outputs.shape[1] == 1, f"Mode mask clouds expects 1 channel output image found {test_outputs.shape}"
            if test_outputs.shape[1] == 3:    
                test_outputs_categorical = test_outputs[:,1] > threshold
                probs_water_pr_curve = test_outputs[:, 1]
            else:
                test_outputs_categorical = test_outputs[:, 0] > threshold
                probs_water_pr_curve = test_outputs[:, 0]
        else:
            assert test_outputs.shape[1] == num_class, f"Mode normal expects {num_class} channel output image found {test_outputs.shape}"
            test_outputs_categorical = torch.argmax(test_outputs, dim=1).long()
            probs_water_pr_curve = test_outputs[:, 1]

        # Ground truth version v1 is 1 channel tensor 3-classes (B, 1, H, W)  {0: invalid, 1: land, 2: water, 3: cloud}
        # Ground truth version v2 is 2 channel tensor 2-classes (B, 2, H, W) [{0: invalid, 1: land, 2: cloud}, {0: invalid, 1: land, 2: water}]
        if ground_truth.shape[1] > 1:
            # Version v2
            assert mask_clouds, f"Expected ground truth of one band found {ground_truth.shape}"
            water_ground_truth = ground_truth[:, 1] # (batch_size, H, W)
            invalids = (water_ground_truth == 0).to(test_outputs_categorical.device) # (batch_size, H, W)
            ground_truth_outputs = torch.clone(water_ground_truth).to(test_outputs_categorical.device)
        else:
            # Version v1
            ground_truth_outputs = torch.clone(ground_truth[:, 0].to(test_outputs_categorical.device))
            # Save invalids to discount
            invalids = ground_truth_outputs == 0

            # Do not consider cloud pixels for metrics
            if mask_clouds:
                invalids |= (ground_truth_outputs == 3)

        ground_truth_outputs[invalids] = 1 # (batch_size, H, W)
        ground_truth_outputs -= 1
        
        # Set invalids in pred to zero
        test_outputs_categorical[invalids] = 0  # (batch_size, H, W)

        confusions_batch = compute_confusions(ground_truth_outputs, test_outputs_categorical,
                                              num_class=num_class, remove_class_zero=False)
        # confusions_batch is (batch_size, num_class, num_class)

        # Discount invalids
        inv_substract = torch.sum(invalids, dim=(1, 2)).to(confusions_batch.device) # (batch_size, )

        confusions_batch[:, 0, 0] -= inv_substract
        confusions.extend(confusions_batch.tolist())
        
        # Thresholded version for precision recall curves
        # Set clouds to land
        test_outputs_categorical_thresh = torch.zeros(ground_truth_outputs.shape, dtype=torch.long,
                                                      device=ground_truth_outputs.device)

        ground_truth_outputs[ground_truth_outputs == 2] = 0

        results = []
        valids = ~invalids
        
        # thresholds_water sorted from high to low
        for thr in thresholds_water:
            # keep invalids in pred to zero
            test_outputs_categorical_thresh[valids & (probs_water_pr_curve > thr)] = 1

            confusions_batch = compute_confusions(ground_truth_outputs,
                                                  test_outputs_categorical_thresh, num_class=2, remove_class_zero=False)   # [batch_size, 2, 2]

            # Discount invalids
            confusions_batch[:, 0, 0] -= torch.sum(invalids.to(confusions_batch.device))

            results.append(confusions_batch.numpy())

        # results is [len(thresholds), batch_size, 2, 2]
        confusions_thresh.append(np.stack(results))

    confusions_thresh = np.concatenate(confusions_thresh, axis=1)
    
    iou = calculate_iou(confusions, labels=label_names)
    recall = calculate_recall(confusions, labels=label_names)
    
    out_dict = {
        'confusions': confusions,
        'confusions_thresholded': confusions_thresh,
        'thresholds': thresholds_water,
        'iou': iou,
        "recall": recall
    }
    
    if plot:
        plot_metrics(out_dict, label_names)
    
    return out_dict



def group_confusion(confusions:torch.Tensor, cems_code:np.ndarray,fun:Callable,
                   label_names:List[str]) ->List[Dict[str, Any]]:
    CMs = OrderedDict({c:[] for c in sorted(np.unique(cems_code))})

    for code, cms in zip(cems_code,confusions):
        CMs[code].append(cms)
    
    data_out = []
    for k, v in CMs.items():
        ious = fun(v, label_names)
        ious["code"] = k
        data_out.append(ious)
    
    return data_out

def compute_metrics_v2(dataloader:torch.utils.data.dataloader.DataLoader,
                       pred_fun: Callable, thresholds_water:Optional[np.array]=None,
                       threshold_water:float=.5, threshold_clouds=.5,
                       plot=False, mask_clouds:bool=True) -> Dict:
    """
    Run inference on a dataloader and compute metrics for that data
    
    Args:
        dataloader: pytorch Dataloader for test set
        pred_fun: function to perform inference using a model
        thresholds_water: list of threshold for precision/recall curves
        threshold_water: threshold of water to compute the confusion matrix
        threshold_clouds: threshold of clouds to compute the confusion matrix
        plot: flag for calling plot method with metrics
        mask_clouds: if True this will compute the confusion matrices for "land" and "water" classes only masking
            the cloudy pixels. If mask_clouds the output confusion matrices will be (n_samples, 2, 2) otherwise they will
            be (n_samples, 3, 3)

        
        returns: dictionary of metrics
    """
    confusions = []
    
    # Sort thresholds from high to low
    if not thresholds_water:
        thresholds_water = [0, 1e-3, 1e-2] + np.arange(0.05, .96, .05).tolist() + [.99, .995, .999]
        thresholds_water = np.array(thresholds_water)

    # thresholds_water sorted from high to low
    thresholds_water = np.sort(thresholds_water)
    thresholds_water = thresholds_water[-1::-1]
    confusions_thresh = []

    # This is constant: we're using this class convention to compute the PR curve
    if mask_clouds:
        num_class, label_names = 2, ["land", "water"]
    else:
        num_class, label_names = 3, ["land", "water", "cloud"]

    for i, batch in tqdm(enumerate(dataloader), total=int(len(dataloader.dataset)/dataloader.batch_size)):
        test_inputs, ground_truth = batch["image"], batch["mask"]

        test_outputs = pred_fun(test_inputs)
      
        test_outputs_categorical = test_outputs[:,1] > threshold_water
        probs_water_pr_curve = test_outputs[:, 1]
        water_ground_truth = ground_truth[:, 1] # (batch_size, H, W)
        invalids = (water_ground_truth == 0).to(test_outputs_categorical.device) # (batch_size, H, W)
        ground_truth_outputs = torch.clone(water_ground_truth).to(test_outputs_categorical.device)
        
        if not mask_clouds:
            assert test_outputs.shape[1] == 2, f"Mode mask clouds expects 2 channel output image found {test_outputs.shape}"
            test_outputs_categorical[test_outputs[:, 0] > threshold_clouds] = 2
            water_ground_truth[ground_truth[:, 0] == 2] = 3 # (batch_size, H, W)
            invalids |= (ground_truth[:, 0] == 0).to(test_outputs_categorical.device)

        # Ground truth version v2 is 2 channel tensor 2-classes (B, 2, H, W) [{0: invalid, 1: land, 2: cloud}, {0: invalid, 1: land, 2: water}]

        ground_truth_outputs[invalids] = 1 # (batch_size, H, W)
        ground_truth_outputs -= 1
        
        # Set invalids in pred to zero
        test_outputs_categorical[invalids] = 0  # (batch_size, H, W)

        confusions_batch = compute_confusions(ground_truth_outputs, test_outputs_categorical,
                                              num_class=num_class, remove_class_zero=False)
        # confusions_batch is (batch_size, num_class, num_class)

        # Discount invalids
        inv_substract = torch.sum(invalids, dim=(1, 2)).to(confusions_batch.device) # (batch_size, )

        confusions_batch[:, 0, 0] -= inv_substract
        confusions.extend(confusions_batch.tolist())
        
        # Thresholded version for precision recall curves
        # Set clouds to land
        test_outputs_categorical_thresh = torch.zeros(ground_truth_outputs.shape, dtype=torch.long,
                                                      device=ground_truth_outputs.device)

        ground_truth_outputs[ground_truth_outputs == 2] = 0

        results = []
        valids = ~invalids
        
        # thresholds_water sorted from high to low
        for thr in thresholds_water:
            # keep invalids in pred to zero
            test_outputs_categorical_thresh[valids & (probs_water_pr_curve > thr)] = 1

            confusions_batch = compute_confusions(ground_truth_outputs,
                                                  test_outputs_categorical_thresh, num_class=2, remove_class_zero=False)   # [batch_size, 2, 2]

            # Discount invalids
            confusions_batch[:, 0, 0] -= torch.sum(invalids.to(confusions_batch.device))

            results.append(confusions_batch.numpy())

        # results is [len(thresholds), batch_size, 2, 2]
        confusions_thresh.append(np.stack(results))

    confusions_thresh = np.concatenate(confusions_thresh, axis=1)
    
    iou = calculate_iou(confusions, labels=label_names)
    recall = calculate_recall(confusions, labels=label_names)
    
    out_dict = {
        'confusions': confusions,
        'confusions_thresholded': confusions_thresh,
        'thresholds': thresholds_water,
        'iou': iou,
        "recall": recall
    }
    
    if plot:
        plot_metrics(out_dict, label_names)
    
    return out_dict


def compute_positives(ground_truth_outputs: torch.Tensor,
                       water_outputs_categorical: torch.Tensor, convert_targets:bool = True) -> torch.Tensor:
    
    """
    Computes FP, FN and TP given a pair of ground truth and water/land predictions 
    
    """
    
    ground_truth = ground_truth_outputs.clone().type(torch.uint8) 
    
    if convert_targets:
        ground_truth = convert_targets_to_v1(ground_truth).squeeze(0)
    
    invalids = ground_truth == 0  # (batch_size, H, W) gpu
    cloudy = ground_truth == 3

    # Set invalids in pred to zero

    water_outputs = water_outputs_categorical.clone().type(torch.uint8) 
    water_outputs[invalids] = 0  # (batch_size, H, W)
    
    FP = (water_outputs == 2) & (ground_truth == 1) #preds water is land
    FN = (water_outputs == 1) & (ground_truth == 2) #preds land is water
    TP = (water_outputs == 2) & (ground_truth == 2) # pred ok!
    
    positives = torch.zeros(size = ground_truth.shape)
    positives[invalids] = 4
    positives[cloudy] = 4
    positives[FP] = 1
    positives[FN] = 2
    positives[TP] = 3

    return positives.squeeze(0)
        




import datetime
from dataclasses import dataclass
from typing import Dict, Set


@dataclass
class MapDataSource:
    """
    Contains information about a data source

    @field path Resource path, starting one directory above "worldfloods"
    @field last_modified Best-effort last-modified date for the asset
    """
    path: str
    last_modified: datetime.datetime

    def __hash__(self):
        return self.path.__hash__()

@dataclass
class MapData:
    """
    Represents known data for a subset of the map separated by type of data

    @field metadata Set of MapDataSources with JSON metadata
    @field satellite_images Dictionary of Sets of MapDataSources with satellite image GeoTIF data, one set per data source
    @field floodmaps Set of MapDataSources containing shape files that detail flood-related geographic information
    @field cloudmasks Set of MapDataSources with GeoTIF cloud masks
    """
    metadata: Set[MapDataSource]
    satellite_images: Dict
    floodmaps: Set[MapDataSource]
    cloudmasks: Set[MapDataSource]
    
    def __init__(self, min_lat, min_lon, max_lat, max_lon, index_list):
       self.metadata = set()
       self.satellite_images = {}
       self.floodmaps = set()
       self.cloudmasks = set()
       for row in range(min_lat, max_lat): 
           for col in range(min_lon, max_lon):
               for descriptor in index_list[(row + 90) * 360 + col + 180]:
                   ds = MapDataSource(path=descriptor["path"], last_modified=descriptor["last_modified"])
                   if descriptor["type"] == "meta":
                       self.metadata.add(ds)
                   elif descriptor["type"] == "satellite_image":
                       if descriptor["provider_id"] not in self.satellite_images:
                           self.satellite_images[descriptor["provider_id"]] = set([ds])
                       else:
                           self.satellite_images[descriptor["provider_id"]].add(ds)
                   elif descriptor["type"] == "floodmap":
                       self.floodmaps.add(ds)
                   elif descriptor["type"] == "cloudmask":
                       self.cloudmasks.add(ds)
                   else:
                       raise Exception("Invalid data descriptor in index")




from setuptools import setup, find_packages
import codecs
import os.path

with open("README.md", "r") as fh:
    long_description = fh.read()


def parse_requirements_file(filename):
    with open(filename, encoding="utf-8") as fid:
        requires = [l.strip() for l in fid.readlines() if l]
    return requires


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

# Optional Packages
# See https://godatadriven.com/blog/a-practical-guide-to-using-setup-py/
EXTRAS = {
    "dev": ["black", "isort", "pylint", "flake8", "pyprojroot"],
    "tests": [
        "pytest",
    ],
    "docs": [
        "furo==2020.12.30b24",
        "nbsphinx==0.8.1",
        "nb-black==1.0.7",
        "matplotlib==3.3.3",
        "sphinx-copybutton==0.3.1",
    ],
}

setup(name="ml4floods",
      version=get_version("ml4floods/__init__.py"),
      author="SpaceML-org",
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=find_packages(".", exclude=["tests"]),
      package_data={
        "ml4floods" : ["models/configurations/*.json",
                       "data/configuration/train_test_split_extra_dataset.json",
                       "data/configuration/train_test_split_original_dataset.json"]  # Add json files from configuration dirs.
      },
      description="Machine learning models for end-to-end flood extent segmentation.",
      install_requires=parse_requirements_file("requirements.txt"),
      extras_require=EXTRAS,
      keywords=["floods pytorch machine-learning earth"],
)



import os.path

import matplotlib.axes
import rasterio
from rasterio import plot as rasterioplt
import rasterio.windows
from matplotlib import colors
import matplotlib.patches as mpatches
import numpy as np
from typing import Union, Optional, List, Tuple
from ml4floods.data.worldfloods.configs import BANDS_S2, BANDS_L8
from ml4floods.models.model_setup import get_channel_configuration_bands
from ml4floods.data.worldfloods import configs
from ml4floods.data import utils
from matplotlib.patches import Patch
import geopandas as gpd

COLORS_WORLDFLOODS = np.array(configs.COLORS_WORLDFLOODS)

COLORS_WORLDFLOODS_V1_1 = np.array([[0, 0, 0], # invalid
                               [139, 64, 0], # land
                               [0, 0, 139], # water
                               [220, 220, 220]], # cloud
                              dtype=np.float32) / 255

INTERPRETATION_WORLDFLOODS = ["invalid", "land", "water", "cloud"]
INTERPRETATION_INVLANDWATER = ["invalid", "land", "water"]

COLORS_WORLDFLOODS_PERMANENT = np.array([[0, 0, 0], # 0: invalid
                                         [139, 64, 0], # 1: land
                                         [237, 0, 0], # 2: flood_water
                                         [220, 220, 220], # 3: cloud
                                         [0, 0, 139], # 4: permanent_water
                                         [60, 85, 92]], # 5: seasonal_water
                                        dtype=np.float32) / 255

INTERPRETATION_WORLDFLOODS_PERMANENT = ["invalid", "land", "flood water", "cloud", "permanent water", "seasonal water"]


def get_cmap_norm_colors(color_array, interpretation_array):
    cmap_categorical = colors.ListedColormap(color_array)
    norm_categorical = colors.Normalize(vmin=-.5,
                                        vmax=color_array.shape[0]-.5)
    patches = []
    for c, interp in zip(color_array, interpretation_array):
        patches.append(mpatches.Patch(color=c, label=interp))
    
    return cmap_categorical, norm_categorical, patches


def _read_data(filename:str,
               bands_rasterio:Optional[List[int]]=None,
               window:Optional[rasterio.windows.Window]=None,
               size_read:Optional[int]=None) -> Tuple[np.ndarray, rasterio.Affine]:
    """
    Reads data from filename with rasterio possibly from the pyramids if `size_read` is not None.
    Returns a tuple with the data read and the affine transformation.

    Args:
        filename:
        bands_rasterio:
        window:
        size_read:

    Returns:
        (C, H, W) array and affine transformation
    """

    with utils.rasterio_open_read(filename) as rst:
        if size_read is not None:
            if bands_rasterio is None:
                n_bands = rst.count
            else:
                n_bands = len(bands_rasterio)

            if window is None:
                shape = rst.shape
            else:
                shape = window.height, window.width

            if (size_read >= shape[0]) and (size_read >= shape[1]):
                out_shape = (n_bands, )+ shape
                input_output_factor = 1
            elif shape[0] > shape[1]:
                out_shape = (n_bands, size_read, int(round(shape[1]/shape[0] * size_read)))
                input_output_factor = shape[0]  / size_read # > 1
            else:
                out_shape = (n_bands, int(round(shape[0] / shape[1] * size_read)), size_read)
                input_output_factor = shape[1] / size_read # > 1
        else:
            out_shape = None
            input_output_factor = None

        output = rst.read(bands_rasterio, window=window, out_shape=out_shape)
        transform = rst.transform if window is None else rasterio.windows.transform(window, rst.transform)

    if input_output_factor is not None:
        transform = rasterio.Affine(transform.a * input_output_factor, transform.b, transform.c,
                                    transform.d, transform.e * input_output_factor, transform.f)

    return output, transform


def get_image_transform(array_or_file:Union[str, np.ndarray],
                        transform:Optional[rasterio.Affine]=None,
                        bands:List[int]=None,
                        window:Optional[rasterio.windows.Window]=None,
                        size_read:Optional[int]=None) -> Tuple[np.ndarray, rasterio.Affine]:
    """
    Reads certain bands and window from `array_or_file`. If `array_or_file` is a file with `size_read` we can read
    from the pyramids of the data to speed up plotting.

    Args:
        array_or_file: array or file to read the data.
        transform: if `array_or_file` is a `np.array`, this current affine transform of it.
        bands: 0-based bands to read.
        window: `rasterio.windows.Window` to read
        size_read: if `array_or_file` is a string, this will be the max size of height and width. It is used to read
        from the pyramids  of the file.

    Returns:
        (C, H, W) array and affine transformation

    """

    if bands is not None:
        bands_rasterio = [b+1 for b in bands]
    else:
        bands_rasterio = None

    if isinstance(array_or_file, str):
        return _read_data(array_or_file, bands_rasterio, window=window, size_read=size_read)

    if hasattr(array_or_file, "cpu"):
        array_or_file = array_or_file.cpu()

    array_or_file = np.array(array_or_file)
    if window is None:
        window_slices = (slice(None), slice(None), slice(None))
    else:
        window_slices = (slice(None),) + window.toslices()

    output = array_or_file[bands, ...]
    output = output[window_slices]
    if transform is not None:
        transform = transform if window is None else rasterio.windows.transform(window, transform)

    return output, transform


def plot_rgb_image(input: Union[str, np.ndarray], transform:Optional[rasterio.Affine]=None,
                   window:Optional[rasterio.windows.Window]=None,
                   max_clip_val:Optional[float]=3000.,
                   min_clip_val:Optional[float]=0.,
                   channel_configuration:str="all",
                   collection_name:str="S2",
                   size_read:Optional[int]=None,
                   **kwargs):
    """
    Plot bands B4, B3, B2 of a Sentinel-2 image. Input could be an array or a str. Values are clipped to 3000
    (it assumes the image has values in [0, 10_000] -> https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2 )

    Tip: Use `size_read` to read from the image pyramid if input is a COG GeoTIFF to speed up the reading.

    Args:
        input: str of array with (C, H, W) configuration.
        transform: geospatial transform if input is an array
        window: window to read from the input
        max_clip_val: value to clip the the input for visualization
        min_clip_val: value to clip the the input for visualization
        channel_configuration: Expected bands of the inputs
        collection_name: S2 or Landsat
        size_read: max size to read. Use this to read from the overviews of the image.
        **kwargs: extra args for rasterio.plot.show

    Returns:
        ax : matplotlib Axes
            Axes with plot.

    """

    # Take into account channel_configuration if input is a np.Array
    assert collection_name in {"Landsat", "S2"}, f"Collection name {collection_name} not implemented"
    if not isinstance(input, str):
        ibands = get_channel_configuration_bands(channel_configuration,
                                                 collection_name=collection_name)
        if collection_name == "S2":
            band_names_current_image = [BANDS_S2[iband] for iband in ibands]
        else:
            band_names_current_image = [BANDS_L8[iband] for iband in ibands]
    else:
        with rasterio.open(input) as rst:
            band_names_current_image = list(rst.descriptions)
        # band_names_current_image = BANDS_S2 if collection_name == "S2" else BANDS_L8

    bands = [band_names_current_image.index(b) for b in ["B4", "B3", "B2"]]
    image, transform = get_image_transform(input, transform=transform, bands=bands, window=window,
                                           size_read=size_read)

    if max_clip_val is not None:
        min_clip_val = 0 if min_clip_val is None else min_clip_val
        image = np.clip((image-min_clip_val)/(max_clip_val - min_clip_val), 0, 1)

    return rasterioplt.show(image, transform=transform, **kwargs)

COLORS_FLOODMAP = {"water": "#0010F6",
                   # "cloud": "#CFCFCF",
                   "cloud": "#4B4B4B",
                   "flood_trace": "#F66F00",
                   "flood-trace": "#F66F00",
                   "water-post-flood": "#FF09E3",
                   "water-pre-flood": "#0010F6",
                   "cloud-pre-flood": "#2B2B2B",
                   "permanent": "#021d5c", # for vectorized JRC Permanent Water Layer
                   "seasonal": "#33aba7",
                   "area_imaged": "#EE0000",
                   "area_imaged-pre-flood": "#AA0000"}

def plot_floodmap(floodmap:Union[str, gpd.GeoDataFrame], legend:bool=True,
                  ax:Optional[matplotlib.axes.Axes]=None,
                  figsize:Tuple[int,int]=(10,10)) -> matplotlib.axes.Axes:
    """
    Plot the vectorized floodmaps as they are produced by inference.py script

    Args:
        floodmap: GeoDataFrame or path to geojson file. The function expects that the floodmap have columns "geometry" and "class".
                 Values of columns class are mapped to COLORS_FLOODMAP
        legend (bool): add legend to the plot
        ax:
        figsize:

    Returns:
           Axes with plot.
    """
    if isinstance(floodmap, str):
        _, ext = os.path.splitext(floodmap)
        if ext == ".geojson":
            floodmap = utils.read_geojson_from_gcp(floodmap)
        else:
            floodmap = gpd.read_file(floodmap)

    floodmap_color = floodmap["class"].apply(lambda x: COLORS_FLOODMAP[x])

    polygons_plot_boundary = (floodmap["class"] == "area_imaged") | (floodmap["class"] == "area_imaged-pre-flood")

    if not polygons_plot_boundary.all():
        ax = floodmap[~polygons_plot_boundary].plot(color=floodmap_color[~polygons_plot_boundary], figsize=figsize, ax=ax)

    # Plot area imaged
    if polygons_plot_boundary.any():
        ax = floodmap[polygons_plot_boundary].boundary.plot(color=floodmap_color[polygons_plot_boundary], ax=ax)

    if legend:
        legend_elements = [Patch(facecolor=COLORS_FLOODMAP[c], label=c) for c in floodmap["class"].unique()]
        ax.legend(handles=legend_elements)
    return ax

def plot_swirnirred_image(input: Union[str, np.ndarray],
                          transform:Optional[rasterio.Affine]=None,
                          window:Optional[rasterio.windows.Window]=None,
                          max_clip_val: Optional[float] = 3000.,
                          min_clip_val: Optional[float] = 0.,
                          channel_configuration="all",
                          collection_name="S2",
                          size_read:Optional[int]=None,
                          **kwargs) -> matplotlib.axes.Axes:
    """
    Plot bands B11, B8, B4 of a Sentinel-2 image. Input could be an array or a str. Values are clipped to 3000
    (it assumes the image has values in [0, 10_000] -> https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2 )

    Tip: Use `size_read` to read from the image pyramid if input is a COG GeoTIFF to speed up the reading.

    Args:
        input: str of array with (C, H, W) configuration.
        transform: geospatial transform if input is an array
        window: window to read from the input
        max_clip_val: value to clip the input for visualization
        min_clip_val: value to clip the input for visualization
        channel_configuration: Expected bands of the inputs
        size_read: max size to read. Use this to read from the overviews of the image.
        collection_name: S2 or Landsat
        **kwargs: extra args for rasterio.plot.show

    Returns:
        ax : matplotlib Axes
            Axes with plot.

    """
    assert collection_name in {"Landsat", "S2"}, f"Collection name {collection_name} not implemented"
    if not isinstance(input, str):
        ibands = get_channel_configuration_bands(channel_configuration,
                                                 collection_name=collection_name)
        if collection_name == "S2":
            band_names_current_image = [BANDS_S2[iband] for iband in ibands]
        else:
            band_names_current_image = [BANDS_L8[iband] for iband in ibands]
    else:
        with rasterio.open(input) as rst:
            band_names_current_image = list(rst.descriptions)

    if collection_name == "S2":
        bands = [band_names_current_image.index(b) for b in ["B11", "B8", "B4"]]
    else:
        bands = [band_names_current_image.index(b) for b in ["B6", "B5", "B4"]]

    image, transform = get_image_transform(input, transform=transform, bands=bands, window=window,
                                           size_read=size_read)

    if max_clip_val is not None:
        min_clip_val = 0 if min_clip_val is None else min_clip_val
        image = np.clip((image-min_clip_val)/(max_clip_val - min_clip_val), 0, 1)

    return rasterioplt.show(image, transform=transform, **kwargs)


def plots_preds_v1(prediction: Union[str, np.ndarray],transform:Optional[rasterio.Affine]=None,
                   window:Optional[rasterio.windows.Window]=None, legend=True,
                   size_read:Optional[int]=None,
                   **kwargs) -> matplotlib.axes.Axes:
    """
    Prediction expected to be {0: land, 1: water, 2: cloud}
    Args:
        prediction: (1, H, W) tensor or path to tiff tile with codes {0: land, 1: water, 2: cloud}
        transform:
        window:
        legend:
        size_read:
        **kwargs:

    Returns:
         ax : matplotlib Axes
            Axes with plot.

    """
    prediction, transform = get_image_transform(prediction,transform=transform, bands=[0], window=window,
                                                size_read=size_read)
    prediction_show = prediction[0] + 1
    cmap_preds, norm_preds, patches_preds = get_cmap_norm_colors(configs.COLORS_WORLDFLOODS,
                                                                 INTERPRETATION_WORLDFLOODS)

    ax = rasterioplt.show(prediction_show, transform=transform, cmap=cmap_preds, norm=norm_preds,
                          interpolation='nearest',**kwargs)

    if legend:
        ax.legend(handles=patches_preds,
                  loc='upper right')

    return ax


def plots_preds_v2(prediction: Union[str, np.ndarray],transform:Optional[rasterio.Affine]=None,
                   window:Optional[rasterio.windows.Window]=None, legend:bool=True,
                   size_read:Optional[int]=None,
                   **kwargs):
    """
    Prediction expected binary (H, W). This function plots only the land/water mask
    Args:
        prediction: (H, W) binary mask
        transform: geotransform
        window: window to read
        legend: plot legend
        size_read:
        **kwargs:

    Returns:

    """
    prediction, transform = get_image_transform(prediction, transform=transform, bands=[0], window=window,
                                                size_read=size_read)
    prediction_show = prediction + 1
    cmap_preds, norm_preds, patches_preds = get_cmap_norm_colors(configs.COLORS_WORLDFLOODS_INVLANDWATER,
                                                                 INTERPRETATION_INVLANDWATER)

    ax = rasterioplt.show(prediction_show, transform=transform, cmap=cmap_preds, norm=norm_preds,
                          interpolation='nearest',**kwargs)

    if legend:
        ax.legend(handles=patches_preds,
                  loc='upper right')

    return ax

def plot_gt_v1(target: Union[str, np.ndarray], transform:Optional[rasterio.Affine]=None,
               window:Optional[rasterio.windows.Window]=None,
               legend=True, size_read:Optional[int]=None, **kwargs):
    """
    ground truth `target` expected to be {0: invalid: 1: land, 2: water, 3: cloud}
    Args:
        target:
        transform:
        window:
        legend:
        size_read:
        **kwargs:

    Returns:

    """

    target, transform = get_image_transform(target,transform=transform, bands=[0], window=window,
                                            size_read=size_read)
    target = target[0]
    cmap_preds, norm_preds, patches_preds = get_cmap_norm_colors(configs.COLORS_WORLDFLOODS,
                                                                 INTERPRETATION_WORLDFLOODS)

    ax = rasterioplt.show(target, transform=transform, cmap=cmap_preds, norm=norm_preds,
                          interpolation='nearest', **kwargs)

    if legend:
        ax.legend(handles=patches_preds,
                  loc='upper right')

    return ax

def plot_gt_v2(target: Union[str, np.ndarray], transform:Optional[rasterio.Affine]=None,
               window:Optional[rasterio.windows.Window]=None,
               legend=True, size_read:Optional[int]=None, **kwargs):
    """
    ground truth `target` expected to be 2 channel image with the following code:
        [{0: invalid: 1: land, 2: cloud}, {0:invalid, 1:land, 2: water}]

    We use the invalid values of the land/water mask

    Args:
        target:
        transform:
        window:
        legend:
        size_read:
        **kwargs:

    Returns:

    """

    target, transform = get_image_transform(target,transform=transform, bands=[0, 1], window=window,
                                            size_read=size_read)
    clear_clouds = target[0]
    land_water = target[1]

    v1gt = land_water.copy() # {0: invalid, 1: land, 2: water}
    v1gt[clear_clouds == 2] = 3

    cmap_preds, norm_preds, patches_preds = get_cmap_norm_colors(configs.COLORS_WORLDFLOODS,
                                                                 INTERPRETATION_WORLDFLOODS)

    ax = rasterioplt.show(v1gt, transform=transform, cmap=cmap_preds, norm=norm_preds,
                          interpolation='nearest', **kwargs)

    if legend:
        ax.legend(handles=patches_preds,
                  loc='upper right')

    return ax


def gt_v1_with_permanent_water(gt: np.ndarray, permanent_water: np.ndarray) -> np.ndarray:
    """ Permanent water taken from: https://developers.google.com/earth-engine/datasets/catalog/JRC_GSW1_3_YearlyHistory"""
    gt[(gt == 2) & (permanent_water == 3)] = 4  # set as permanent_water
    gt[(gt == 2) & (permanent_water == 2)] = 5  # set as seasonal water

    return gt


def plot_gt_v1_with_permanent(target: Union[str, np.ndarray], permanent: Optional[Union[str, np.ndarray]]=None,
                              transform:Optional[rasterio.Affine]=None,
                              window:Optional[rasterio.windows.Window]=None, legend=True,
                              size_read:Optional[int]=None,
                              **kwargs):
    bands = [0]
    target, transform = get_image_transform(target, transform=transform, bands=bands, window=window,
                                            size_read=size_read)
    target= target[0]
    if permanent is not None:
        permanent, _ = get_image_transform(permanent, transform=transform, bands=bands, window=window,
                                           size_read=min(target.shape))
        permanent = permanent[0]
        target = gt_v1_with_permanent_water(target, permanent)

    cmap_gt, norm_gt, patches_gt = get_cmap_norm_colors(COLORS_WORLDFLOODS_PERMANENT, INTERPRETATION_WORLDFLOODS_PERMANENT)

    ax = rasterioplt.show(target,transform=transform, cmap=cmap_gt, norm=norm_gt,
                          interpolation='nearest', **kwargs)

    if legend:
        ax.legend(handles=patches_gt,
                  loc='upper right')

    return ax


def plot_s2_and_confusions(input: Union[str, np.ndarray], positives: np.ndarray ,title:Optional[str] = None,
                           transform:Optional[rasterio.Affine]=None, channel_configuration = 'all', **kwargs):
    """
    Plots a S2 image and overlapping FP, FN and TP with masked clouds, computed from 
    compute_positives function

    """
    band_names_current_image = [BANDS_S2[iband] for iband in get_channel_configuration_bands(channel_configuration)]
    bands = [band_names_current_image.index(b) for b in ["B11", "B8", "B4"]]
    image = get_image_transform(input, transform=transform, bands=bands)[0]
    image = np.clip((image-0)/(3000 - 0), 0, 1)
    image = np.moveaxis(image,0,-1)
    
    #set invalids and clouds to black, FP white, FN orange and TP blue
    image[positives == 4] = colors.to_rgb('black')
    image[positives == 1] = colors.to_rgb('C9')
    image[positives == 2] = colors.to_rgb('orange')
    image[positives == 3] = colors.to_rgb('blue')

    cmap_colors = ['orange','C9', 'blue']
    cmap = colors.ListedColormap(cmap_colors)
    
    return rasterioplt.show(np.moveaxis(image,-1,0), cmap = cmap, transform = transform, title = title, **kwargs)



import torch
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt

from ml4floods.preprocess.worldfloods import normalize
from ml4floods.data.worldfloods.configs import COLORS_WORLDFLOODS, CHANNELS_CONFIGURATIONS, BANDS_S2


def compute_uncertainties(dataloader, p_pred_fun, d_pred_fun, num_class, config, num_samples=10):
    """
    Run dropout during inference on a dataloader and compute uncertainty metrics for that data, plot union and intersection of 'focus_label' predictions onto an image for visual uncertainty inspection
    
    Args:
    :param dataloader: pytorch Dataloader for test set
    :param pred_fun: function to perform inference using a model
    :param num_class: number of classes
    :param focus_label: name of class we're interested in finding uncertainty around
    :param num_samples: number of samples to take from network to calculate uncertainty from
    :return: None
    """
    for i, batch in tqdm(enumerate(dataloader), total=int(len(dataloader.dataset) / dataloader.batch_size)):
        test_inputs, ground_truth_outputs = batch['image'], batch['mask'].squeeze(1)
        compute_uncertainties_for_image_pair(test_inputs, ground_truth_outputs, p_pred_fun, d_pred_fun, num_samples, num_class, config)
        
        


def compute_uncertainties_for_image_pair(inputs, targets, p_pred_fun, d_pred_fun, num_samples, num_class, config, denorm=True):
    """
    Runs probabilistic inference on image pair - takes multiple samples from network and computes uncertainty maps, generates a plot of those maps

    :param inputs: Tensor (BxCxHxW)
    :param targets: Tensor (1xCxHxW)
    :param p_pred_fun: probabilistic function - every sample from function varies
    :param d_pred_fun: deterministic function - every sample from function is the same
    :param num_samples: number of samples to take using the probabilistic function
    :param num_class: number of classes in target
    :param config: experiment config
    :return: None
    """
    with torch.no_grad():
        prediction_samples = np.zeros([num_samples] + list(targets.shape))
        prediction_logits = np.zeros([num_samples] + [1, num_class] + list(targets.squeeze().shape))

        for s in range(num_samples):
            sample_logits = p_pred_fun(inputs)

            prediction_logits[s] = sample_logits.cpu().numpy()

            sample_prediction = torch.argmax(sample_logits, dim=1).long()

            prediction_samples[s] = sample_prediction.cpu().numpy()

        samples = prediction_samples
        images = inputs.cpu().numpy()
        gts = targets.cpu().numpy()
        predictions = torch.argmax(d_pred_fun(inputs), dim=1).long().cpu().numpy()

        if denorm:
            optical = s1_to_unnormed_rgb(images, config)[0]
        else:
            optical = s1_to_rgb(images, config)[0]

        gt = mask_to_rgb(gts[0])
        prediction = mask_to_rgb(predictions[0] + 1)
        diff = (predictions[0] - gts[0])

        water_prob = water_probability(samples)
        water_bound = water_bounds(samples)
        water_ent = water_entropy(samples)
        class_var = variance_map(samples)

        fig, ax = plt.subplots(2, 4, figsize=(40, 20))

        # Optical Image
        ax[0, 0].imshow(optical)
        ax[0, 0].set_title('Input', fontweight="bold", size=40)
        # Ground Truth
        ax[0, 1].imshow(gt, interpolation='nearest')
        ax[0, 1].set_title('Ground Truth', fontweight="bold", size=40)
        # Prediction
        ax[0, 2].imshow(prediction, interpolation='nearest')
        ax[0, 2].set_title('Prediction', fontweight="bold", size=40)
        # Diff Map
        ax[0, 3].imshow(diff)
        ax[0, 3].set_title('Difference Map', fontweight="bold", size=40)
        # Uncertainty 1: Water probability
        ax[1, 0].imshow(water_prob, cmap='Blues')
        ax[1, 0].set_title('Water Probability', fontweight="bold", size=40)
        # Uncertainty 2: Water bounds
        ax[1, 1].imshow(water_bound, cmap='Blues')
        ax[1, 1].set_title('Water Bounds', fontweight="bold", size=40)
        # Uncertainty 3: Water Entropy
        ax[1, 2].imshow(water_ent)
        ax[1, 2].set_title('Water Entropy', fontweight="bold", size=40)
        # Uncertainty 4: All Class Variance (TODO: Replace with Mutual Information)
        ax[1, 3].imshow(class_var)
        ax[1, 3].set_title('Class Variance', fontweight="bold", size=40)

        plt.show()


def water_probability(samples):
    water_only = to_binary_segmentation(samples, 1)
    total_image = np.sum(water_only, axis=0)[0]
    return total_image


def water_bounds(samples):
    water_only = to_binary_segmentation(samples, 1)
    total_image = np.sum(water_only, axis=0)

    water_union = total_image.copy()
    water_union[water_union != 0] = 0.5

    water_intersection = total_image.copy()
    water_intersection[water_intersection != len(samples)] = 0
    water_intersection[water_intersection != 0] = 0.5

    water_bounds = water_union - water_intersection

    water_bounds[water_intersection == 0.5] = 1

    return water_bounds[0]


def water_entropy(samples):
    water_only = to_binary_segmentation(samples, 1)
    total_image = np.sum(water_only, axis=0)
    p = (total_image / len(samples))[0]

    entropy_image = np.zeros_like(p)
    for i in range(p.shape[0]):
        for j in range(p.shape[1]):
            if p[i, j] != 0:
                entropy_image[i, j] = -(p[i, j] * np.log(p[i, j]))

    return entropy_image


def variance_map(samples):
    var_samples = np.var(samples, axis=0)[0]
    return var_samples


def to_binary_segmentation(samples, class_value):
    # Extract single class maps from each sample
    binary_only = []
    for s in samples:
        binary_image = s.copy()
        binary_image[binary_image != class_value] = 0
        binary_only.append(binary_image)

    binary_only = np.array(binary_only)

    return binary_only


def s1_to_unnormed_rgb(image, config):
    """
    :param image: normalised input image
    :param config: experiment config
    :return: unnormalized rgb image
    """
    mean, std = normalize.get_normalisation("bgr")  # B, R, G!
    mean = mean[np.newaxis]
    std = std[np.newaxis]

    # Find the RGB indexes within the S2 bands
    bands_read_names = [BANDS_S2[i] for i in
                        CHANNELS_CONFIGURATIONS[config["model_params"]["hyperparameters"]['channel_configuration']]]
    bands_index_rgb = [bands_read_names.index(b) for b in ["B4", "B3", "B2"]]

    model_input_rgb_npy = image[:, bands_index_rgb].transpose(0, 2, 3, 1) * std[..., -1::-1] + mean[..., -1::-1]
    model_input_rgb_npy = np.clip(model_input_rgb_npy / 3000., 0., 1.)
    return model_input_rgb_npy


def s1_to_rgb(image, config):

    # Find the RGB indexes within the S2 bands
    bands_read_names = [BANDS_S2[i] for i in CHANNELS_CONFIGURATIONS[config["model_params"]["hyperparameters"]['channel_configuration']]]
    bands_index_rgb = [bands_read_names.index(b) for b in ["B4", "B3", "B2"]]

    model_input_rgb_npy = image[:, bands_index_rgb].transpose(0, 2,3,1) 
    model_input_rgb_npy = np.clip(model_input_rgb_npy / 3000., 0., 1.)
    return model_input_rgb_npy


def mask_to_rgb(mask, values=[0,1,2,3], colors_cmap=COLORS_WORLDFLOODS):
    """
    Given a 2D mask it assign each value of the mask the corresponding color
    :param mask:
    :param values:
    :param colors_cmap:
    :return:
    """
    assert len(values) == len(
        colors_cmap), f"Values and colors should have same length {len(values)} {len(colors_cmap)}"
    mask_return = np.zeros(mask.shape[:2] + (3,), dtype=np.uint8)
    colores = np.array(np.round(colors_cmap * 255), dtype=np.uint8)
    for i, c in enumerate(colores):
        mask_return[mask == values[i], :] = c
    return mask_return



from ml4floods.preprocess.worldfloods import normalize as wf_normalization
import ml4floods.preprocess.transformations as transformations
from ml4floods.preprocess.tiling import WindowSlices, load_windows, save_windows
from glob import glob
from ml4floods.data.worldfloods.configs import CHANNELS_CONFIGURATIONS
from ml4floods.data.worldfloods.dataset import WorldFloodsDatasetTiled
from ml4floods.data.worldfloods.lightning import WorldFloodsDataModule
from ml4floods.models.utils.configuration import AttrDict
import pytorch_lightning as pl
import os
import json
from typing import Dict, List, Callable, Tuple, Optional
from ml4floods.preprocess.worldfloods import prepare_patches
from ml4floods.models.config_setup import load_json, get_filesystem
import warnings


def filenames_train_test_split(bucket_name:Optional[str], train_test_split_file:str) -> Dict[str, Dict[str, List[str]]]:
    """
    Read train test split json file from remote if needed.

    Args:
        bucket_name:
        train_test_split_file:

    Returns:

    """
    if bucket_name or train_test_split_file.startswith("gs://"):
        if not train_test_split_file.startswith("gs://"):
            train_test_split_file = f"gs://{bucket_name}/{train_test_split_file}"
        filenames_train_test = load_json(train_test_split_file)

        # Preprend the bucket name to the files
        for splitname, split in filenames_train_test.items():
            for foldername, listoftiffs in split.items():
                for idx,tiffname in enumerate(listoftiffs):
                    if not tiffname.startswith("gs://"):
                        tiffname = f"gs://{bucket_name}/{tiffname}"
                        # assert fs.exists(tiffname), f"File {tiffname} does not exists"
                        listoftiffs[idx] = tiffname
    else:
        with open(train_test_split_file, "r") as fh:
            filenames_train_test = json.load(fh)

    return filenames_train_test


def validate_worldfloods_data(path_to_splits:str,
                              split_folders:List[str] = ["train", "val", "test"],
                              folders_to_test:List[str] = ["S2", "gt"],
                              verbose:bool=True):
    """
    Check the structure of the data follows the expected convention
    Args:
        path_to_splits: path where train/test/val splits are.
        split_folders: Split folders to check
        folders_to_test: products to test (i.e. this could be S2, PERMANENTWATERJRC, floodmaps,...)
        verbose: prints success

    Raises:
        FileNotFoundError if something goes wrong

    """
    if not os.path.exists(path_to_splits):
        raise FileNotFoundError(f" Path to splits folder not found {path_to_splits}")

    for isplit in split_folders:
        folder_split = os.path.join(path_to_splits, isplit)
        if not os.path.exists(folder_split):
            raise FileNotFoundError(f"Splits folder not found {folder_split}")

        filenames = None
        for foldername in folders_to_test:
            folder_path = os.path.join(folder_split, foldername)
            if not os.path.exists(folder_path):
                raise FileNotFoundError(f" Folder not found {folder_path}")
            filenames_folder = [os.path.basename(os.path.splitext(f)[0]) for f in sorted(glob(os.path.join(folder_path, "*")))]
            if len(filenames_folder) == 0:
                raise FileNotFoundError(f"Folder {folder_path} does not have any files on it")
            if filenames is None:
                filenames = filenames_folder
            else:
                if filenames != filenames_folder:
                    raise FileNotFoundError(f"Different files {filenames} {filenames_folder}")
    if verbose:
        print("Data downloaded follows the expected format")


def process_filename_train_test(train_test_split_file:Optional[str]="gs://ml4cc_data_lake/2_PROD/2_Mart/worldfloods_v1_0/train_test_split.json",
                                input_folder:str="S2",
                                target_folder:str="gt",
                                bucket_id:Optional[str]=None, path_to_splits:Optional[str]=None,
                                download:Optional[Dict[str,bool]]=None) -> Dict[str,Dict[str, List[str]]]:
    """
    The train_test_split_file contains which files go to the train/val/test splits. This function validate that
    the content is as expected and that all files referred there exists. Additionally it downloads the data to loca
    if specified.

    Args:
        train_test_split_file:
        input_folder:
        target_folder:
        bucket_id:
        path_to_splits:
        download: e.g. `{"train": True, "val": False, "test": True}` to download train and val data.

    Returns:

    """

    if download is None:
        download = {"train": False, "val": False, "test": False}

    if train_test_split_file:
        filenames_train_test = filenames_train_test_split(bucket_id, train_test_split_file)
    else:
        assert (path_to_splits is not None) and os.path.exists(path_to_splits), \
            f"train_test_split_file not provided and path_to_splits folder {path_to_splits} does not exist"

        print(f"train_test_split_file not provided. We will use the content in the folder {path_to_splits}")
        filenames_train_test = {'train': {target_folder:[], input_folder:[]},
                                'test': {target_folder:[],input_folder:[]},
                                'val': {target_folder:[],input_folder:[]}}

    # loop through the naming splits
    for isplit in ["train", "test", "val"]:
        for foldername in [input_folder, target_folder]:

            # glob files in path_to_splits dir if there're not files in the given split
            if len(filenames_train_test[isplit][foldername]) == 0:
                # get the subdirectory
                assert (path_to_splits is not None) and os.path.exists(path_to_splits), \
                    f"path_to_splits {path_to_splits} doesn't exists or not provided and there're no files in split {isplit} folder {foldername}"

                path_2_glob = os.path.join(path_to_splits, isplit, foldername, "*.tif")
                filenames_train_test[isplit][foldername] = glob(path_2_glob)
                assert len(filenames_train_test[isplit][foldername]) > 0, f"No files found in {path_2_glob}"

        assert len(filenames_train_test[isplit][input_folder]) == len(filenames_train_test[isplit][target_folder]), \
            f"Different number of files in {input_folder} and {target_folder} for split {isplit}: {len(filenames_train_test[isplit][input_folder])} {len(filenames_train_test[isplit][target_folder])}"

        # check correspondence input output files (assert files exists)
        for idx, filename in enumerate(filenames_train_test[isplit][input_folder]):
            fs = get_filesystem(filename)
            assert fs.exists(filename), f"File input: {filename} does not exists"

            filename_target = filenames_train_test[isplit][target_folder][idx]
            assert fs.exists(filename_target), f"File target: {filename_target} does not exists"

            # Download if needed and replace filenames_train_test with the downloaded version
            if filename.startswith("gs://") and download[isplit]:
                assert (path_to_splits is not None) and os.path.exists(path_to_splits), \
                    f"path_to_splits {path_to_splits} doesn't exists or not provided ad requested to download the data"

                for input_target_folder in [input_folder, target_folder]:
                    folder_local = os.path.join(path_to_splits, isplit, input_target_folder)
                    os.makedirs(folder_local, exist_ok=True)
                    basename = os.path.basename(filename)
                    file_dest = os.path.join(folder_local, basename)
                    if not os.path.isfile(file_dest):
                        fs.get_file(filename, file_dest)
                        print(f"Downloaded ({idx}/{len(filenames_train_test[isplit][input_folder])}) {filename}")
                    filenames_train_test[isplit][input_folder][idx] = file_dest

    return filenames_train_test


def get_dataset(data_config) -> pl.LightningDataModule:
    """
    Function to set up dataloaders for model training
    """
    
    # 1. Setup transformations for dataset
    
    train_transform, test_transform = get_transformations(data_config)

    # ======================================================
    # Obtain train/val/test files
    # ======================================================
    download = data_config.get("download")
    if download is None:
        download = {"train": data_config.get("loader_type","local")=="local",
                    "val": data_config.get("loader_type","local")=="local",
                    "test": data_config.get("loader_type","local")=="local"}

    filenames_train_test = process_filename_train_test(data_config.get("train_test_split_file"),
                                                       input_folder=data_config.input_folder,
                                                       target_folder=data_config.target_folder,
                                                       bucket_id=data_config.get("bucket_id"),
                                                       path_to_splits=data_config.get("path_to_splits"),
                                                       download=download)

    filter_windows_attr = data_config.get("filter_windows", None)
    if filter_windows_attr is not None and filter_windows_attr.get("apply", False):
        filter_windows_config = filter_windows_fun(data_config.filter_windows.version, data_config.train_test_split_file,
                                                   threshold_clouds=data_config.filter_windows.threshold_clouds,
                                                   local_destination_dir=data_config.path_to_splits)
    else:
        filter_windows_config = None

    # CREATE DATAMODULE
    datamodule = WorldFloodsDataModule(
        filenames_train_test=filenames_train_test,
        input_folder=data_config.input_folder,
        target_folder=data_config.target_folder,
        train_transformations=train_transform,
        test_transformations=test_transform,
        bands=CHANNELS_CONFIGURATIONS[data_config.channel_configuration],
        num_workers=data_config.num_workers,
        window_size=data_config.window_size,
        batch_size=data_config.batch_size,
        filter_windows= filter_windows_config
    )
    datamodule.setup()

    print("train", datamodule.train_dataset.__len__(), " tiles")
    print("val", datamodule.val_dataset.__len__(), " tiles")
    print("test", datamodule.test_dataset.__len__(), " tiles")
    
    return datamodule

def filter_windows_fun(data_version:str, train_test_split_file:str, local_destination_dir:Optional[str]=None, threshold_clouds=.5) -> Callable:
    """
    Returns a function to filter the windows in the  WorldFloodsDatasetTiled dataset. This is used for pre-filtering
    the training images to discard patches with high cloud content.

    Args:
        data_version: "v1" or "v2"
        local_destination_dir: local destination to save the json file with the windows to use
        threshold_clouds: threshold to use to filter the window (window will be filtered if it has more than threshold_clouds
        clouds)

    Returns:
        function to filter windows

    """
    if local_destination_dir is not None:
        if train_test_split_file:
            split_name = "_"+os.path.basename(train_test_split_file)
        else:
            split_name =".json"
        #windows_file = os.path.join(local_destination_dir, f"windows_{data_version}.json")
        windows_file = os.path.join(local_destination_dir, f"windows{split_name}")
    else:
        windows_file = None

    def filter_windows(tiledDataset: WorldFloodsDatasetTiled) -> List[WindowSlices]:
        if windows_file and os.path.exists(windows_file):
            selected_windows = load_windows(windows_file)
        else:
            if data_version == "v1":
                selected_windows =  prepare_patches.filter_windows_v1(tiledDataset,
                                                                      threshold_clouds=threshold_clouds)
            elif data_version == "v2":
                selected_windows = prepare_patches.filter_windows_v2(tiledDataset,
                                                                     threshold_clouds=threshold_clouds)
            else:
                raise NotImplementedError(f"Unknown ground truth version {data_version} expected v1 or v2")

            if windows_file:
                save_windows(selected_windows, windows_file)

        return selected_windows

    return filter_windows


def get_transformations(data_config) -> Tuple[Callable, Callable]:
    """
    Function to generate transformations object to pass to dataloader
    TODO: Build from config instead of using default values
    """

    train_transform = [
        transformations.InversePermuteChannels(),
        transformations.RandomRotate90(always_apply=True, p=0.5),
        transformations.Flip(always_apply=True, p=0.5)]
    
    if "train_transformation" not in data_config:
        warnings.warn("Train transformation not found in data config. Assume normalize is True")
        data_config["train_transformation"] = AttrDict({"normalize": True})

    channel_mean = None
    if data_config.train_transformation.normalize:
        channel_mean, channel_std = wf_normalization.get_normalisation(data_config.channel_configuration)
        train_transform.append(transformations.Normalize(
            mean=channel_mean,
            std=channel_std,
            max_pixel_value=1))

    train_transform.extend([
        transformations.PermuteChannels(),
        transformations.ToTensor(),
    ])

    train_transform = transformations.Compose(train_transform)

    if "test_transformation" not in data_config:
        warnings.warn("Test transformation not found in data config. Assume normalize is True")
        data_config["test_transformation"] = AttrDict({"normalize": True})
    
    if data_config.test_transformation.normalize:
        if channel_mean is None:
            channel_mean, channel_std = wf_normalization.get_normalisation(data_config.channel_configuration)

        test_transform = [
        transformations.InversePermuteChannels(), 
        transformations.Normalize(
            mean=channel_mean, 
            std=channel_std, 
            max_pixel_value=1),
        transformations.PermuteChannels(),
        transformations.ToTensor(),
        ]
        test_transform = transformations.Compose(test_transform)
    else:
        test_transform = transformations.ToTensor()
    
    return train_transform, test_transform
